"""
Constructed Molecule
====================

"""

import logging
import rdkit.Chem.AllChem as rdkit

from .molecule import Molecule
from ..atoms import AtomInfo
from ..bonds import BondInfo
from .utilities import get_bond_info_atom_ids, sort_bond_atoms_by_id

logger = logging.getLogger(__name__)


class ConstructedMolecule(Molecule):
    """
    Represents constructed molecules.

    Examples
    --------
    *Initialization*

    A :class:`.ConstructedMolecule` is initialized from a
    :class:`.TopologyGraph`, which is typically initialized from some
    :class:`.BuildingBlock` instances.

    .. code-block:: python

        import stk

        bb1 = stk.BuildingBlock('NCCCN', [stk.PrimaryAminoFactory()])
        bb2 = stk.BuildingBlock(
            smiles='O=CC(C=O)CC=O',
            functional_groups=[stk.AldehydeFactory()],
        )
        tetrahedron = stk.cage.FourPlusSix((bb1, bb2))
        cage1 = stk.ConstructedMolecule(tetrahedron)

    A :class:`ConstructedMolecule` may be used to construct other
    :class:`ConstructedMolecule` instances, though you will probably
    have to convert it to a :class:`.BuildingBlock` first

    .. code-block:: python

        benzene = stk.BuildingBlock('c1ccccc1')
        cage_complex = stk.host_guest.Complex(
            host=stk.BuildingBlock.init_from_molecule(cage1),
            guest=benzene,
        )
        cage_complex = stk.ConstructedMolecule(host_guest_complex)

    Obviously, the initialization of the :class:`.ConstructedMolecule`
    depends mostly on the specifics of the :class:`.TopologyGraph`
    used, and the documentation of those classes should be examined
    for more examples.

    """

    def __init__(self, topology_graph):
        """
        Initialize a :class:`.ConstructedMolecule`.

        Parameters
        ----------
        topology_graph : :class:`.TopologyGraph`
            The topology graph of the constructed molecule.

        """

        construction_result = topology_graph.construct()
        super().__init__(
            atoms=construction_result.get_atoms(),
            bonds=construction_result.get_bonds(),
            position_matrix=construction_result.get_position_matrix(),
        )
        self._atom_infos = construction_result.get_atom_infos()
        self._bond_infos = construction_result.get_bond_infos()
        self._num_building_blocks = {
            building_block:
                topology_graph.get_num_building_block(building_block)
            for building_block in topology_graph.get_building_blocks()
        }

    @classmethod
    def init(
        cls,
        atoms,
        bonds,
        position_matrix,
        atom_infos,
        bond_infos,
        num_building_blocks,
    ):
        """
        Initialize a :class:`.ConstructedMolecule` from its components.

        Parameters
        ----------
        atoms : :class:`tuple` of :class:`.Atom`
            The atoms of the molecule.

        bond : :class:`tuple` of :class:`.Bond`
            The bonds of the molecule.

        position_matrix : :class:`numpy.ndarray`
            A ``(n, 3)`` position matrix of the molecule.

        atom_infos : :class:`tuple` of :class:`.AtomInfo`
            The atom infos of the molecule.

        bond_infos : :class:`tuple` of :class:`.BondInfo`
            The bond infos of the molecule.

        num_building_blocks : :class:`dict`
            Maps each building block of the constructed molecule to
            the number of times it is present in it.

        Returns
        -------
        :class:`.ConstructedMolecule`
            The constructed molecule.

        """

        molecule = cls.__new__(cls)
        Molecule.__init__(molecule, atoms, bonds, position_matrix)
        molecule._atom_infos = atom_infos
        molecule._bond_infos = bond_infos
        molecule._num_building_blocks = dict(num_building_blocks)
        return molecule

    def clone(self):
        clone = super().clone()
        clone._atom_infos = self._atom_infos
        clone._bond_infos = self._bond_infos
        clone._num_building_blocks = dict(self._num_building_blocks)
        return clone

    def get_building_blocks(self):
        """
        Yield the building blocks of the constructed molecule.

        Building blocks are yielded in an order based on their
        position in the constructed molecule. For two topologically
        equivalent constructed molecules, but with different building
        blocks, equivalently positioned building blocks will be
        yielded at the same time.

        Yields
        ------
        :class:`.Molecule`
            A building block of the constructed molecule.

        """

        yield from self._num_building_blocks

    def get_num_building_block(self, building_block):
        """
        Get the number of times `building_block` is present.

        Parameters
        ----------
        building_block : :class:`.Molecule`
            The building block whose frequency in the constructed
            molecule is desired.

        Returns
        -------
        :class:`int`
            The number of times `building_block` was used in the
            construction of the constructed molecule.

        """

        return self._num_building_blocks[building_block]

    def get_atom_infos(self, atom_ids=None):
        """
        Yield data about atoms in the molecule.

        Parameters
        ----------
        atom_ids : :class:`iterable` of :class:`int`, optional
            The ids of atoms whose data is desired. If ``None``,
            data on all atoms will be yielded. Can be a single
            :class:`int`, if data on a single atom is desired.

        Yields
        ------
        :class:`.AtomInfo`
            Data about an atom.

        """

        if atom_ids is None:
            atom_ids = range(len(self._atoms))
        elif isinstance(atom_ids, int):
            atom_ids = (atom_ids, )

        for atom_id in atom_ids:
            yield self._atom_infos[atom_id]

    def get_bond_infos(self):
        """
        Yield data about bonds in the molecule.

        Yields
        ------
        :class:`.BondInfo`
            Data about a bond.

        """

        yield from self._bond_infos

    def _with_canonical_atom_ordering(self):
        ordering = rdkit.CanonicalRankAtoms(self.to_rdkit_mol())
        id_map = {
            new_id: atom.get_id()
            for new_id, atom in zip(ordering, self._atoms)
        }
        super()._with_canonical_atom_ordering()
        atom_map = {
            old_id: self._atoms[new_id]
            for old_id, new_id in enumerate(ordering)
        }
        old_atom_infos = self._atom_infos

        def get_atom_info(atom):
            info = old_atom_infos[id_map[atom.get_id()]]
            return AtomInfo(
                atom=atom,
                building_block=info.get_building_block(),
                building_block_id=info.get_building_block_id(),
            )

        def get_bond_info(info):
            return BondInfo(
                bond=sort_bond_atoms_by_id(
                    info.get_bond().with_atoms(atom_map)
                ),
                building_block=info.get_building_block(),
                building_block_id=info.get_building_block_id(),
            )

        self._atom_infos = tuple(map(get_atom_info, self._atoms))
        self._bond_infos = tuple(sorted(
            map(get_bond_info, self._bond_infos),
            key=get_bond_info_atom_ids,
        ))
        return self



"""
Random Topology Graph
=====================

"""

import numpy as np

from .mutator import MoleculeMutator
from ...records import MutationRecord
from ....molecule_records import MoleculeRecord


class RandomTopologyGraph(MoleculeMutator):
    """
    Changes topology graphs at random.

    Examples
    --------
    *Constructed Molecule Mutation*

    .. code-block:: python

        import stk

        # Create a molecule which is to be mutated.
        bb1 = stk.BuildingBlock('NCCN', [stk.PrimaryAminoFactory()])
        bb2 = stk.BuildingBlock(
            smiles='O=CCC(=O)CC=O',
            functional_groups=[stk.AldehydeFactory()],
        )
        cage = stk.MoleculeRecord(
            topology_graph=stk.cage.FourPlusSix((bb1, bb2),
        )

        # Create functions which replace the topology graph.
        replacement_funcs = (
            lambda graph:
                stk.cage.TwoPlusThree(graph.get_building_blocks()),
            lambda graph:
                stk.cage.EightPlusTwelve(graph.get_building_blocks()),
            lambda graph:
                stk.cage.TwentyPlusThirty(graph.get_building_blocks()),
        )

        # Create the mutator.
        random_topology = stk.RandomTopologyGraph(replacement_funcs)

        # Mutate a molecule.
        mutation_record1 = random_topology.mutate(cage)

        # Mutate the molecule a second time.
        mutation_record2 = random_topology.mutate(cage)

    """

    def __init__(
        self,
        replacement_funcs,
        name='RandomTopologyGraph',
        random_seed=None,
    ):
        """
        Initialize a :class:`.RandomTopology` instance.

        Parameters
        ----------
        replacement_funcs : :class:`tuple` of :class:`callable`
            Each :class:`callable` takes a single parameter, a
            :class:`.TopologyGraph`, and returns the
            :class:`.TopologyGraph` which should replace it.

        name : :class:`str`, optional
            A name to help identify the mutator instance.

        random_seed : :class:`bool`, optional
            The random seed to use.

        """

        self._replacement_funcs = replacement_funcs
        self._name = name
        self._generator = np.random.RandomState(random_seed)

    def mutate(self, record):
        replacement_func = self._generator.choice(
            a=self._replacement_funcs,
        )
        replacement = replacement_func(record.get_topology_graph())
        return MutationRecord(
            molecule_record=MoleculeRecord(replacement),
            mutator_name=self._name,
        )



"""
Three Plus Six
==============

"""

import numpy as np

from ..cage import Cage
from ..vertices import _LinearCageVertex, _NonLinearCageVertex
from ...topology_graph import Edge


class ThreePlusSix(Cage):
    """
    Represents a cage topology graph.

    See :class:`.Cage` for more details and examples.

    """

    _x = 1
    _vertex_prototypes = (
        _NonLinearCageVertex(0, [-2*_x, -_x*np.sqrt(3), 0], False),
        _NonLinearCageVertex(1, [2*_x, -_x*np.sqrt(3), 0], False),
        _NonLinearCageVertex(2, [0, _x*np.sqrt(3), 0], False),

        _LinearCageVertex(3, [0, -2*_x*np.sqrt(3), _x], False),
        _LinearCageVertex(4, [0, -2*_x*np.sqrt(3), -_x], False),

        _LinearCageVertex(5, [2*_x, 0, _x], False),
        _LinearCageVertex(6, [2*_x, 0, -_x], False),

        _LinearCageVertex(7, [-2*_x, 0, _x], False),
        _LinearCageVertex(8, [-2*_x, 0, -_x], False),
    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[3], _vertex_prototypes[0]),
        Edge(1, _vertex_prototypes[3], _vertex_prototypes[1]),

        Edge(2, _vertex_prototypes[4], _vertex_prototypes[0]),
        Edge(3, _vertex_prototypes[4], _vertex_prototypes[1]),

        Edge(4, _vertex_prototypes[5], _vertex_prototypes[1]),
        Edge(5, _vertex_prototypes[5], _vertex_prototypes[2]),

        Edge(6, _vertex_prototypes[6], _vertex_prototypes[1]),
        Edge(7, _vertex_prototypes[6], _vertex_prototypes[2]),

        Edge(8, _vertex_prototypes[7], _vertex_prototypes[0]),
        Edge(9, _vertex_prototypes[7], _vertex_prototypes[2]),

        Edge(10, _vertex_prototypes[8], _vertex_prototypes[0]),
        Edge(11, _vertex_prototypes[8], _vertex_prototypes[2]),

    )

    _num_windows = 5
    _num_window_types = 2



"""
Value Database
==============

.. toctree::
    :maxdepth: 2

    Value MongoDB <stk.databases.mongo_db.value>

"""


class ValueDatabase:
    """
    Abstract base class for storing of molecular property values.

    Examples
    --------
    *Subclass Implementation*

    The source of any of the subclasses, listed in
    :mod:`molecule_value_database <.databases.value>`,
    can serve as good examples.

    *Iterating Through Entries in Database*

    The :meth:`.get_all` method of a molecule database instance can
    be used to iterate through entries, the keys of which can be used
    to access values.

    .. code-block:: python

        client = pymongo.MongoClient()
        molecule_db = stk.MoleculeMongoDb(client)

        value_db = stk.ValueMongoDb(
            mongo_client=client,
            collection='atom_counts',
        )

        for molecule in molecule_db.get_all():
            try:
                value = value_db.get(molecule)
            except KeyError:
                # In case molecule is not in value_db.
                pass

    """

    def put(self, molecule, value):
        """
        Put a value into the database.

        Parameters
        ----------
        molecule : :class:`.Molecule`
            The molecule which is associated with the `value`.

        value : :class:`object`
            Some value associated with `molecule`.

        Returns
        -------
        None : :class:`NoneType`

        """

        raise NotImplementedError()

    def get(self, molecule):
        """
        Get the stored value for `molecule`.

        Parameters
        ----------
        molecule : :class:`.Molecule`
            The molecule whose value is to be retrieved from the
            database.

        Returns
        -------
        :class:`object`
            The value associated with `molecule`.

        Raises
        ------
        :class:`KeyError`
            If `molecule` is not found in the database.

        """

        raise NotImplementedError()



from stk.molecular import Atom, Bond, AtomInfo, BondInfo


def to_atom(id, json):
    return Atom(id, json[0], json[1])


def to_bond(atoms, json):
    return Bond(
        atom1=atoms[json[0]],
        atom2=atoms[json[1]],
        order=json[2],
        periodicity=tuple(json[3]),
    )


def to_atom_info(building_blocks, atom, json):
    return AtomInfo(
        atom=atom,
        building_block=(
            building_blocks[json[0]]
            if json[0] is not None
            else None
        ),
        building_block_id=json[1],
    )


def to_bond_info(building_blocks, bond, json):
    return BondInfo(
        bond=bond,
        building_block=(
            building_blocks[json[0]]
            if json[0] is not None
            else None
        ),
        building_block_id=json[1],
    )



import numpy as np
from scipy.spatial.distance import euclidean

from ..topology_graph import Vertex


class _CycleVertex(Vertex):
    """
    Represents a vertex in a macrocycle.

    """

    def __init__(self, id, position, flip, angle):
        """
        Initialize a :class:`._CycleVertex` instance.

        Parameters
        ----------
        id : :class:`int`
            The id of the vertex.

        position : :class:`tuple` of :class:`float`
            The position of the vertex.

        flip : :class:`bool`
            If ``True``, the orientation of building blocks placed by
            the vertex will be flipped.

        angle : :class:`float`
            The position of the vertex along the cycle, specified by
            the `angle`.

        """

        super().__init__(id, position)
        self._flip = flip
        self._angle = angle

    def clone(self):
        clone = super().clone()
        clone._flip = self._flip
        clone._angle = self._angle
        return clone

    def get_flip(self):
        """
        Return ``True`` if the vertex flips building blocks it places.

        Returns
        -------
        :class:`bool`
            ``True`` if the vertex flips building blocks it places.

        """

        return self._flip

    def place_building_block(self, building_block, edges):
        assert (
            building_block.get_num_functional_groups() == 2
        ), (
            f'{building_block} needs to have exactly 2 functional '
            'groups but has '
            f'{building_block.get_num_functional_groups()}.'
        )
        building_block = building_block.with_centroid(
            position=self._position,
            atom_ids=building_block.get_placer_ids(),
        )
        fg0, fg1 = building_block.get_functional_groups()
        fg0_position = building_block.get_centroid(
            atom_ids=fg0.get_placer_ids(),
        )
        fg1_position = building_block.get_centroid(
            atom_ids=fg1.get_placer_ids(),
        )
        return building_block.with_rotation_between_vectors(
            start=fg1_position - fg0_position,
            target=[-1 if self._flip else 1, 0, 0],
            origin=self._position,
        ).with_rotation_about_axis(
            angle=self._angle-(np.pi/2),
            axis=np.array([0, 0, 1]),
            origin=self._position,
        ).get_position_matrix()

    def map_functional_groups_to_edges(self, building_block, edges):
        fg0_position = building_block.get_centroid(
            atom_ids=next(
                building_block.get_functional_groups()
            ).get_placer_ids(),
        )

        def fg0_distance(edge):
            return euclidean(edge.get_position(), fg0_position)

        edge0 = min(edges, key=fg0_distance)
        return {
            0: edge0.get_id(),
            1: (
                edges[1].get_id()
                if edge0 is edges[0]
                else edges[0].get_id()
            ),
        }

    def __str__(self):
        return (
            f'Vertex(id={self._id}, '
            f'position={self._position.tolist()}, '
            f'flip={self._flip}, '
            f'angle={self._angle})'
        )



"""
M4L8
====

"""

from ..cage import Cage
from ..vertices import _NonLinearCageVertex, _LinearCageVertex
from ...topology_graph import Edge


class M4L8(Cage):
    """
    Represents a cage topology graph.

    Metal building blocks with four functional groups are
    required for this topology.

    Ligand building blocks with two functional groups are required for
    this topology.

    When using a :class:`dict` for initialization, a
    :class:`.BuildingBlock` needs to be assigned to each of the
    following numbers:

        | metals: (0, 1, 2, 3)
        | ligands: (4, 5, 6, 7, 8, 9, 10, 11)

    See :class:`.Cage` for more details and examples.

    """

    _vertex_prototypes = (
        _NonLinearCageVertex(0, [1, 0, 0]),
        _NonLinearCageVertex(1, [0, 1, 0]),
        _NonLinearCageVertex(2, [-1, 0, 0]),
        _NonLinearCageVertex(3, [0, -1, 0]),

        _LinearCageVertex(4, [1, 1, 0.5], False),
        _LinearCageVertex(5, [1, 1, -0.5], False),

        _LinearCageVertex(6, [1, -1, 0.5], False),
        _LinearCageVertex(7, [1, -1, -0.5], False),

        _LinearCageVertex(8, [-1, -1, 0.5], False),
        _LinearCageVertex(9, [-1, -1, -0.5], False),

        _LinearCageVertex(10, [-1, 1, 0.5], False),
        _LinearCageVertex(11, [-1, 1, -0.5], False),
    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[0], _vertex_prototypes[4]),
        Edge(1, _vertex_prototypes[0], _vertex_prototypes[5]),
        Edge(2, _vertex_prototypes[0], _vertex_prototypes[6]),
        Edge(3, _vertex_prototypes[0], _vertex_prototypes[7]),

        Edge(4, _vertex_prototypes[1], _vertex_prototypes[4]),
        Edge(5, _vertex_prototypes[1], _vertex_prototypes[5]),
        Edge(6, _vertex_prototypes[1], _vertex_prototypes[10]),
        Edge(7, _vertex_prototypes[1], _vertex_prototypes[11]),

        Edge(8, _vertex_prototypes[2], _vertex_prototypes[10]),
        Edge(9, _vertex_prototypes[2], _vertex_prototypes[11]),
        Edge(10, _vertex_prototypes[2], _vertex_prototypes[8]),
        Edge(11, _vertex_prototypes[2], _vertex_prototypes[9]),

        Edge(12, _vertex_prototypes[3], _vertex_prototypes[8]),
        Edge(13, _vertex_prototypes[3], _vertex_prototypes[9]),
        Edge(14, _vertex_prototypes[3], _vertex_prototypes[6]),
        Edge(15, _vertex_prototypes[3], _vertex_prototypes[7]),
    )

    _num_windows = 2
    _num_window_types = 1



"""
Functional Group
================

.. toctree::
    :maxdepth: 2

    Alcohol <stk.molecular.functional_groups.functional_groups.alcohol>
    Aldehyde <\
stk.molecular.functional_groups.functional_groups.aldehyde\
>
    Alkene <stk.molecular.functional_groups.functional_groups.alkene>
    Alkyne <stk.molecular.functional_groups.functional_groups.alkyne>
    Amide <stk.molecular.functional_groups.functional_groups.amide>
    Boronic Acid <\
stk.molecular.functional_groups.functional_groups.boronic_acid\
>
    Bromo <stk.molecular.functional_groups.functional_groups.bromo>
    Carboxylic Acid <\
stk.molecular.functional_groups.functional_groups.carboxylic_acid\
>
    Dibromo <stk.molecular.functional_groups.functional_groups.dibromo>
    Difluoro <\
stk.molecular.functional_groups.functional_groups.difluoro\
>
    Diol <stk.molecular.functional_groups.functional_groups.diol>
    Fluoro <stk.molecular.functional_groups.functional_groups.fluoro>
    Generic Functional Group <\
stk.molecular.functional_groups.functional_groups.\
generic_functional_group\
>
    Iodo <stk.molecular.functional_groups.functional_groups.iodo>
    Primary Amino <\
stk.molecular.functional_groups.functional_groups.primary_amino\
>
    Ring Amine <\
stk.molecular.functional_groups.functional_groups.ring_amine\
>
    Secondary Amino <\
stk.molecular.functional_groups.functional_groups.secondary_amino\
>
    Single Atom <\
stk.molecular.functional_groups.functional_groups.single_atom\
>
    Thioacid <\
stk.molecular.functional_groups.functional_groups.thioacid\
>
    Thiol <stk.molecular.functional_groups.functional_groups.thiol>


Functional groups define which atoms of a :class:`.BuildingBlock` are
modified during :class:`.ConstructedMolecule` construction, and which
are used to position it.
The class of a :class:`.FunctionalGroup`
affects which :class:`.Reaction` can be used with it.
See the abstract base class :class:`.FunctionalGroup` for more
information.

"""


class FunctionalGroup:
    """
    An abstract base class for functional groups.

    It is used to give access to atoms of :class:`.BuildingBlock`
    molecules which are modified during
    :class:`.ConstructedMolecule` construction, as well as specify
    which atoms of the building block should be used for positioning.

    The most common reason you would want to implement a new
    :class:`.FunctionalGroup` subclass, is because you want to
    customize the construction of a :class:`.ConstructedMolecule`.
    Specifically, you want to modify a specific set of atoms in a
    :class:`.BuildingBlock` when doing construction, and you want
    to modify them in a specific way. You will usually accompany the
    creation of the new :class:`.FunctionalGroup` subclass with the
    creation of a new :class:`.Reaction` subclass, which will perform
    the custom modification on the atoms held by your new
    :class:`.FunctionalGroup`
    subclass. Finally, a new :class:`.ReactionFactory` will also be
    created, so that your :class:`.Reaction` subclass instances
    actually get made during construction. Finally, you will pass an
    instance of your :class:`.ReactionFactory` subclass to the
    chosen :class:`.TopologyGraph` you want to make, for example
    :class:`~.polymer.linear.linear.Linear`, and your custom
    modification will take place.

    See Also
    --------
    :mod:`.functional_group_factory`
        Used for automated creation of :class:`.FunctionalGroup`
        instances.

    Notes
    -----
    You might notice that some of the methods of this abstract base
    class are implemented. This is purely for convenience when
    implementing subclasses. The implemented public methods are simply
    default implementations, which can be safely ignored
    or overridden, when implementing subclasses. Any private methods
    are implementation details of these default implementations.

    Examples
    --------
    *Subclass Implementation*

    The source code of the subclasses, listed in
    :mod:`.functional_group`, can serve as good examples.

    """

    def __init__(self, atoms, placers, core_atoms):
        """
        Initialize a :class:`.FunctionalGroup`.

        Parameters
        ----------
        atoms : :class:`tuple` of :class:`.Atom`
            The atoms in the functional group.

        placers : :class:`tuple` of :class:`.Atom`
            The atoms used to calculate the position of the functional
            group.

        core_atoms : :class:`tuple` of :class:`.Atom`
            The atoms of the functional group which also form the core
            of the :class:`.BuildingBlock`. See
            :meth:`.BuildingBlock.get_core_atom_ids`.

        """

        self._atoms = atoms
        self._placers = placers
        self._core_atoms = core_atoms

    def get_atoms(self):
        """
        Yield all the atoms in the functional group.

        Yields
        ------
        :class:`.Atom`
            An atom in the functional group.

        """

        yield from self._atoms

    def get_atom_ids(self):
        """
        Yield the ids of all atoms in the functional group.

        Yields
        ------
        :class:`int`
            The id of an :class:`.Atom`.

        """

        yield from (a.get_id() for a in self._atoms)

    def get_placer_ids(self):
        """
        Yield the ids of *placer* atoms.

        *Placer* atoms are those, which should be used to calculate
        the position of the functional group.

        Yields
        ------
        :class:`int`
            The id of an :class:`.Atom`.

        """

        yield from (a.get_id() for a in self._placers)

    def get_core_atom_ids(self):
        """
        Yield the ids of core atoms held by the functional group.

        Yields
        ------
        :class:`int`
            The id of an :class:`.Atom`.

        See Also
        --------
        :meth:`.BuildingBlock.get_core_atom_ids`

        """

        yield from (a.get_id() for a in self._core_atoms)

    def _with_atoms(self, atom_map):
        """
        Modify the functional group.

        """

        self._atoms = tuple(
            atom_map.get(a.get_id(), a) for a in self._atoms
        )
        self._placers = tuple(
            atom_map.get(a.get_id(), a) for a in self._placers
        )
        self._core_atoms = tuple(
            atom_map.get(a.get_id(), a) for a in self._core_atoms
        )
        return self

    def with_atoms(self, atom_map):
        """
        Return a clone holding different atoms.

        Parameters
        ----------
        atom_map : :class:`dict`
            Maps the id of an atom in the functional group to the new
            atom the clone should hold. If the id of an atom in the
            functional group is not found in `atom_map`, the atom
            will not be replaced in the clone.

        Returns
        -------
        :class:`.FunctionalGroup`
            The clone. Has the same type as the original functional
            group.

        Examples
        --------
        .. code-block:: python

            import stk

            n, h1, h2 = stk.N(0), stk.H(1), stk.H(2)
            amine = stk.PrimaryAmino(
                nitrogen=n,
                hydrogen1=h1,
                hydrogen2=h2,
                bonders=(n, ),
                deleters=(h1, h2),
            )

            n20 = stk.N(20)
            h100 = stk.H(100)

            # fg_clone is a clone of fg, except that instead of holding
            # n, fg_clone holds n20, and instead of holding h1
            # fg_clone holds h100. fg_clone continues to hold h2.
            fg_clone = fg.with_atoms({
                n.get_id(): n20,
                h1.get_id(): h100,
            })


        """

        return self.clone()._with_atoms(atom_map)

    def clone(self):
        """
        Return a clone.

        Returns
        -------
        :class:`.FunctionalGroup`
            A clone. Has the same type as the original functional
            group.

        """

        clone = self.__class__.__new__(self.__class__)
        FunctionalGroup.__init__(
            self=clone,
            atoms=self._atoms,
            placers=self._placers,
            core_atoms=self._core_atoms,
        )
        return clone

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return (
            f'{self.__class__.__name__}('
            f'atoms={self._atoms}, '
            f'placers={self._placers}, '
            f'core_atoms={self._core_atoms}'
            ')'
        )



"""
Four Plus Six 2
===============

"""

from ..cage import Cage
from ..vertices import _LinearCageVertex, _NonLinearCageVertex
from ...topology_graph import Edge


class FourPlusSix2(Cage):
    """
    Represents a cage topology graph.

    See :class:`.Cage` for more details and examples.

    """

    _vertex_prototypes = (
        _NonLinearCageVertex(0, [1, 0, 1]),
        _NonLinearCageVertex(1, [-1, 0, 1]),
        _NonLinearCageVertex(2, [1, 0, -1]),
        _NonLinearCageVertex(3, [-1, 0, -1]),

        _LinearCageVertex(4, [0, -1, 1], False),
        _LinearCageVertex(5, [0, 1, 1], False),
        _LinearCageVertex(6, [0, -1, -1], False),
        _LinearCageVertex(7, [0, 1, -1], False),
    )

    _vertex_prototypes = (
        *_vertex_prototypes,
        _LinearCageVertex.init_at_center(
            id=8,
            vertices=(_vertex_prototypes[0], _vertex_prototypes[2]),
        ),
        _LinearCageVertex.init_at_center(
            id=9,
            vertices=(_vertex_prototypes[1], _vertex_prototypes[3]),
        )
    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[4], _vertex_prototypes[0]),
        Edge(1, _vertex_prototypes[4], _vertex_prototypes[1]),
        Edge(2, _vertex_prototypes[5], _vertex_prototypes[0]),
        Edge(3, _vertex_prototypes[5], _vertex_prototypes[1]),
        Edge(4, _vertex_prototypes[6], _vertex_prototypes[2]),
        Edge(5, _vertex_prototypes[6], _vertex_prototypes[3]),
        Edge(6, _vertex_prototypes[7], _vertex_prototypes[2]),
        Edge(7, _vertex_prototypes[7], _vertex_prototypes[3]),
        Edge(8, _vertex_prototypes[8], _vertex_prototypes[0]),
        Edge(9, _vertex_prototypes[8], _vertex_prototypes[2]),
        Edge(10, _vertex_prototypes[9], _vertex_prototypes[1]),
        Edge(11, _vertex_prototypes[9], _vertex_prototypes[3]),
    )

    _num_windows = 4
    _num_window_types = 2



"""
This module defines general-purpose objects, functions and classes.

Functions, classes, etc. defined here should not depend on any other
part of ``stk``. They must be completely self-sufficient.

"""

import rdkit.Chem.AllChem as rdkit
from rdkit.Geometry import Point3D
import numpy as np
import time
from contextlib import contextmanager
import os
import re
from collections import deque
import tarfile


# Holds the elements Van der Waals radii in Angstroms.
atom_vdw_radii = {
    'Al': 2, 'Sb': 2, 'Ar': 1.88, 'As': 1.85, 'Ba': 2,
    'Be': 2, 'Bi': 2, 'B': 2, 'Br': 1.85, 'Cd': 1.58,
    'Cs': 2, 'Ca': 2, 'C': 1.7, 'Ce': 2, 'Cl': 1.75,
    'Cr': 2, 'Co': 2, 'Cu': 1.4, 'Dy': 2, 'Er': 2,
    'Eu': 2, 'F': 1.47, 'Gd': 2, 'Ga': 1.87, 'Ge': 2,
    'Au': 1.66, 'Hf': 2, 'He': 1.4, 'Ho': 2, 'H': 1.09,
    'In': 1.93, 'I': 1.98, 'Ir': 2, 'Fe': 2, 'Kr': 2.02,
    'La': 2, 'Pb': 2.02, 'Li': 1.82, 'Lu': 2, 'Mg': 1.73,
    'Mn': 2, 'Hg': 1.55, 'Mo': 2, 'Nd': 2, 'Ne': 1.54,
    'Ni': 1.63, 'Nb': 2, 'N': 1.55, 'Os': 2, 'O': 1.52,
    'Pd': 1.63, 'P': 1.8, 'Pt': 1.72, 'K': 2.75, 'Pr': 2,
    'Pa': 2, 'Re': 2, 'Rh': 2, 'Rb': 2, 'Ru': 2, 'Sm': 2,
    'Sc': 2, 'Se': 1.9, 'Si': 2.1, 'Ag': 1.72, 'Na': 2.27,
    'Sr': 2, 'S': 1.8, 'Ta': 2, 'Te': 2.06, 'Tb': 2,
    'Tl': 1.96, 'Th': 2, 'Tm': 2, 'Sn': 2.17, 'Ti': 2,
    'W': 2, 'U': 1.86, 'V': 2, 'Xe': 2.16, 'Yb': 2,
    'Y': 2, 'Zn': 1.29, 'Zr': 2, 'X': 1.0, 'D': 1.0
}

# This dictionary gives easy access to the rdkit bond types.
bond_dict = {
    '1': rdkit.rdchem.BondType.SINGLE,
    'am': rdkit.rdchem.BondType.SINGLE,
    '2': rdkit.rdchem.BondType.DOUBLE,
    '3': rdkit.rdchem.BondType.TRIPLE,
    'ar': rdkit.rdchem.BondType.AROMATIC
}

# A dictionary which matches atomic number to elemental symbols.
periodic_table = {
    1: 'H', 2: 'He', 3: 'Li', 4: 'Be', 5: 'B', 6: 'C',
    7: 'N', 8: 'O', 9: 'F', 10: 'Ne', 11: 'Na', 12: 'Mg',
    13: 'Al', 14: 'Si', 15: 'P', 16: 'S', 17: 'Cl',
    18: 'Ar', 19: 'K', 20: 'Ca', 21: 'Sc', 22: 'Ti',
    23: 'V', 24: 'Cr', 25: 'Mn', 26: 'Fe', 27: 'Co',
    28: 'Ni', 29: 'Cu', 30: 'Zn', 31: 'Ga', 32: 'Ge',
    33: 'As', 34: 'Se', 35: 'Br', 36: 'Kr', 37: 'Rb',
    38: 'Sr', 39: 'Y', 40: 'Zr', 41: 'Nb', 42: 'Mo',
    43: 'Tc', 44: 'Ru', 45: 'Rh', 46: 'Pd', 47: 'Ag',
    48: 'Cd', 49: 'In', 50: 'Sn', 51: 'Sb', 52: 'Te',
    53: 'I', 54: 'Xe', 55: 'Cs', 56: 'Ba', 57: 'La',
    58: 'Ce', 59: 'Pr', 60: 'Nd', 61: 'Pm', 62: 'Sm',
    63: 'Eu', 64: 'Gd', 65: 'Tb', 66: 'Dy', 67: 'Ho',
    68: 'Er', 69: 'Tm', 70: 'Yb', 71: 'Lu', 72: 'Hf',
    73: 'Ta', 74: 'W', 75: 'Re', 76: 'Os', 77: 'Ir',
    78: 'Pt', 79: 'Au', 80: 'Hg', 81: 'Tl', 82: 'Pb',
    83: 'Bi', 84: 'Po', 85: 'At', 86: 'Rn', 87: 'Fr',
    88: 'Ra', 89: 'Ac', 90: 'Th', 91: 'Pa', 92: 'U',
    93: 'Np', 94: 'Pu', 95: 'Am', 96: 'Cm', 97: 'Bk',
    98: 'Cf', 99: 'Es', 100: 'Fm', 101: 'Md', 102: 'No',
    103: 'Lr', 104: 'Rf', 105: 'Db', 106: 'Sg', 107: 'Bh',
    108: 'Hs', 109: 'Mt', 110: 'Ds', 111: 'Rg', 112: 'Cn',
    113: 'Uut', 114: 'Fl', 115: 'Uup', 116: 'Lv',
    117: 'Uus', 118: 'Uuo'
}


class Cell:
    """
    Represents an individual cell in a supercell.

    Parameters
    ----------
    id : :class:`list` of :class:`int`
        A 3 member :class:`list` holding the x, y and z index
        of the cell within the supercell.

    fgs : :class:`dict`
        Maps the fgs in the original unit cell to the
        equivalent fgs in the cell.

    Attributes
    ----------
    id : :class:`numpy.ndarray` of :class:`int`
        A 3 member array holding the x, y and z index
        of the cell within the supercell.

    fgs : :class:`dict`
        Maps the fgs in the original unit cell to the
        equivalent fgs in the cell.

    """

    def __init__(self, id_, fgs):
        self.id = np.array(id_)
        self.fgs = fgs


class ChargedMolError(Exception):
    def __init__(self, mol_file, msg):
        self.mol_file = mol_file
        self.msg = msg


class MolFileError(Exception):
    def __init__(self, mol_file, msg):
        self.mol_file = mol_file
        self.msg = msg


class PopulationSizeError(Exception):
    def __init__(self, msg):
        self.msg = msg


class LazyAttr:
    """
    A descriptor for creating lazy attributes.

    """

    def __init__(self, func):
        self.func = func

    def __get__(self, obj, cls):
        if obj is None:
            return self
        val = self.func(obj)
        setattr(obj, self.func.__name__, val)
        return val


def archive_output():
    """
    Places the ``output`` folder into ``stk_ea_runs``.

    This function assumes that the ``output`` folder is in the current
    working directory.

    Returns
    -------
    None : :class:`NoneType`

    """

    if 'output' not in os.listdir():
        return

    # Make the ``stk_ea_runs`` folder if it does not exist already.
    if 'stk_ea_runs' not in os.listdir():
        os.mkdir('stk_ea_runs')

    # Find out with what number the ``output`` folder should be
    # labelled within ``stk_ea_runs``.
    num = len(os.listdir('stk_ea_runs'))
    new_dir = os.path.join('stk_ea_runs', str(num))
    os.rename('output', new_dir)


def dedupe(iterable, key=None, seen=None):
    """
    Yields items from `iterable` barring duplicates.

    If `seen` is provided it contains elements which are not to be
    yielded at all.

    Parameters
    ----------
    iterable : :class:`iterable`
        An iterable of elements which are to be yielded, only once.

    key : :class:`callable`
        A function which gets applied to every member of `iterable`.
        The return of this function is checked for duplication rather
        than the member itself.

    seen : :class:`set`, optional
        Holds items which are not to be yielded.

    Yields
    ------
    :class:`object`
        Element in `iterable` which is not in `seen` and has not been
        yielded before.

    """

    if seen is None:
        seen = set()
    for x in iterable:
        val = key(x) if key is not None else x
        if val not in seen:
            seen.add(val)
            yield x


def flatten(iterable, excluded_types=None):
    """
    Transforms an nested iterable into a flat one.

    For example

    .. code-block:: python

        [[1,2,3], [[4], [5],[[6]], 7]

    becomes

    .. code-block:: python

        [1,2,3,4,5,6,7]

    If a type is found in `excluded_types` it will not be yielded from.
    For example if ``str`` is in `excluded_types`

    .. code-block:: python

        a = ["abcd", ["efgh"]]

    "abcd" and "efgh" are yielded if `a` is passed to `iterable`. If
    `str` was not in `excluded_types` then "a", "b", "c", "d", "e",
    "f", "g" and "h" would all be yielded individually.

    Parameters
    ----------
    iterable : :class:`iterable`
        The iterable which is to be flattened

    excluded_types : :class:`set`, optional
        Holds container types which are not be flattened.

    Yields
    ------
    :class:`object`
        A nested element of `iterable`.

    """

    if excluded_types is None:
        excluded_types = {str}

    for x in iterable:
        if hasattr(x, '__iter__') and type(x) not in excluded_types:
            yield from flatten(x, excluded_types)
        else:
            yield x


def kabsch(coords1, coords2):
    """
    Return a rotation matrix to minimize dstance between 2 coord sets.

    This is essentially an implementation of the Kabsch algorithm.
    Given two sets of coordinates, `coords1` and `coords2`, this
    function returns a rotation matrix. When the rotation matrix is
    applied to `coords1` the resulting coordinates have their rms
    distance to `coords2` minimized.

    Parameters
    ----------
    coords1 : :class:`numpy.ndarray`
        This array represents a matrix hodling coordinates which need
        to be rotated to minimize their rms distance to coordinates in
        `coords2`. The matrix is n x 3. Each row of the matrix holds
        the x, y and z coordinates of one point, respectively. Here
        ``n`` is the number of points.

    coords2 : :class:`numpy.ndarray`
        This array represents a matrix which holds the coordinates of
        points the distance to which should be minimized. The matrix is
        ``n x 3``. Each row of the matrix holds the x, y and z
        coordinates of one point, respectively. Here ``n`` is the
        number of points.

    Returns
    -------
    :class:`numpy.ndarray`
        A rotation matrix. This will be a ``3 x 3`` matrix.

    References
    ----------
    http://nghiaho.com/?page_id=671
    https://en.wikipedia.org/wiki/Kabsch_algorithm

    """

    h = np.dot(coords1, coords2.T)
    u, s, v = np.linalg.svd(h)

    if int(np.linalg.det(v)) < 0:
        v[:, 2] = -v[:, 2]

    return np.dot(v, u)


def matrix_centroid(matrix):
    """
    Returns the centroid of the coordinates held in `matrix`.

    Parameters
    ----------
    matrix : :class:`np.ndarray`
        A ``n x 3`` matrix. Each row holds the x, y and z
        coordinate of some point, respectively.

    Returns
    -------
    :class:`numpy.ndarray`
        A numpy array which holds the x, y and z
        coordinates of the centroid of the coordinates in `matrix`.

    """

    return np.sum(matrix, axis=0) / len(matrix)


def mol_from_mae_file(mae_path):
    """
    Creates a ``rdkit`` molecule from a ``.mae`` file.

    Parameters
    ----------
    mol2_file : :class:`str`
        The full path of the ``.mae`` file from which an rdkit molecule
        should be instantiated.

    Returns
    -------
    :class:`rdkit.Mol`
        An ``rdkit`` instance of the molecule held in `mae_file`.

    """

    mol = rdkit.EditableMol(rdkit.Mol())
    conf = rdkit.Conformer()

    with open(mae_path, 'r') as mae:
        content = re.split(r'[{}]', mae.read())

    prev_block = deque([''], maxlen=1)
    for block in content:
        if 'm_atom[' in prev_block[0]:
            atom_block = block
        if 'm_bond[' in prev_block[0]:
            bond_block = block
        prev_block.append(block)

    labels, data_block, *_ = atom_block.split(':::')
    labels = [label for label in labels.split('\n')
              if not label.isspace() and label != '']

    data_block = [a.split() for a in data_block.split('\n') if
                  not a.isspace() and a != '']

    for line in data_block:
        line = [word for word in line if word != '"']
        if len(labels) != len(line):
            raise RuntimeError(('Number of labels does'
                                ' not match number of columns'
                                ' in .mae file.'))

        for label, data in zip(labels, line):
            if 'x_coord' in label:
                x = float(data)
            if 'y_coord' in label:
                y = float(data)
            if 'z_coord' in label:
                z = float(data)
            if 'atomic_number' in label:
                atom_num = int(data)

        atom_sym = periodic_table[atom_num]
        atom_coord = Point3D(x, y, z)
        atom_id = mol.AddAtom(rdkit.Atom(atom_sym))
        conf.SetAtomPosition(atom_id, atom_coord)

    labels, data_block, *_ = bond_block.split(':::')
    labels = [label for label in labels.split('\n')
              if not label.isspace() and label != '']
    data_block = [a.split() for a in data_block.split('\n')
                  if not a.isspace() and a != '']

    for line in data_block:
        if len(labels) != len(line):
            raise RuntimeError(('Number of labels does'
                                ' not match number of '
                                'columns in .mae file.'))

        for label, data in zip(labels, line):
            if 'from' in label:
                atom1 = int(data) - 1
            if 'to' in label:
                atom2 = int(data) - 1
            if 'order' in label:
                bond_order = str(int(data))
        mol.AddBond(atom1, atom2, bond_dict[bond_order])

    mol = mol.GetMol()
    mol.AddConformer(conf)
    return mol


def normalize_vector(vector):
    """
    Normalizes the given vector.

    A new vector is returned, the original vector is not modified.

    Parameters
    ----------
    vector : :class:`np.ndarray`
        The vector to be normalized.

    Returns
    -------
    :class:`np.ndarray`
        The normalized vector.

    """

    return np.divide(vector, np.linalg.norm(vector))


def remake(mol):
    """
    Remakes a molecule from scratch.

    Parameters
    ----------
    mol : :class:`rdkit.Mol`
        The molecule to be remade.

    Returns
    -------
    :class:`rdkit.Mol`
        The remade molecule.

    """

    emol = rdkit.EditableMol(rdkit.Mol())
    for atom in mol.GetAtoms():
        new = rdkit.Atom(atom.GetAtomicNum())
        new.SetFormalCharge(atom.GetFormalCharge())
        emol.AddAtom(new)

    for bond in mol.GetBonds():
        emol.AddBond(
            beginAtomIdx=bond.GetBeginAtomIdx(),
            endAtomIdx=bond.GetEndAtomIdx(),
            order=bond.GetBondType()
        )

    m = emol.GetMol()
    m.AddConformer(mol.GetConformer())
    return m


def get_projection(start, target):
    """
    Get the projection of `start` onto `target`.

    """

    return target * np.dot(
        start,
        target,
    ) / np.dot(target, target)


def orthogonal_vector(vector):
    ortho = np.array([0., 0., 0.])
    for m, val in enumerate(vector):
        if not np.allclose(val, 0, atol=1e-8):
            n = (m+1) % 3
            break
    ortho[n] = vector[m]
    ortho[m] = -vector[n]
    return ortho


def rotation_matrix(vector1, vector2):
    """
    Returns a rotation matrix which transforms `vector1` to `vector2`.

    Multiplying `vector1` by the rotation matrix returned by this
    function yields `vector2`.

    Parameters
    ----------
    vector1 : :class:`numpy.ndarray`
        The vector which needs to be transformed to `vector2`.

    vector2 : :class:`numpy.ndarray`
        The vector onto which `vector1` needs to be transformed.

    Returns
    -------
    :class:`numpy.ndarray`
        A rotation matrix which transforms `vector1` to `vector2`.

    References
    ----------
    http://tinyurl.com/kybj9ox
    http://tinyurl.com/gn6e8mz

    """

    # Make sure both inputs are unit vectors.
    vector1 = normalize_vector(vector1)
    vector2 = normalize_vector(vector2)

    # Hande the case where the input and output vectors are equal.
    if np.allclose(vector1, vector2, atol=1e-8):
        return np.identity(3)

    # Handle the case where the rotation is 180 degrees.
    if np.allclose(vector1, np.multiply(vector2, -1), atol=1e-8):
        return rotation_matrix_arbitrary_axis(
            angle=np.pi,
            axis=orthogonal_vector(vector1)
        )

    v = np.cross(vector1, vector2)
    vx = np.array([
        [0, -v[2], v[1]],
        [v[2], 0, -v[0]],
        [-v[1], v[0], 0]
    ])
    s = np.linalg.norm(v)
    c = np.dot(vector1, vector2)
    i = np.identity(3)
    mult_factor = (1-c)/np.square(s)
    return i + vx + np.multiply(np.dot(vx, vx), mult_factor)


def rotation_matrix_arbitrary_axis(angle, axis):
    """
    Returns a rotation matrix of `angle` radians about `axis`.

    Parameters
    ----------
    angle : :class:`float`
        The size of the rotation in radians.

    axis : :class:`numpy.ndarray`
        A 3 element aray which represents a vector. The vector is the
        axis about which the rotation is carried out. Must be of
        unit magnitude.

    Returns
    -------
    :class:`numpy.ndarray`
        A ``3x3`` array representing a rotation matrix.

    """

    a = np.cos(angle/2)
    b, c, d = axis * np.sin(angle/2)

    e11 = np.square(a) + np.square(b) - np.square(c) - np.square(d)
    e12 = 2*(b*c - a*d)
    e13 = 2*(b*d + a*c)

    e21 = 2*(b*c + a*d)
    e22 = np.square(a) + np.square(c) - np.square(b) - np.square(d)
    e23 = 2*(c*d - a*b)

    e31 = 2*(b*d - a*c)
    e32 = 2*(c*d + a*b)
    e33 = np.square(a) + np.square(d) - np.square(b) - np.square(c)

    return np.array([[e11, e12, e13],
                     [e21, e22, e23],
                     [e31, e32, e33]])


def dice_similarity(mol1, mol2, fp_radius=3):
    """
    Return the chemical similarity between two molecules.

    Parameters
    ----------
    mol1 : :class:`.Molecule`
        The first molecule.

    mol2 : :class:`.Molecule`
        The second molecule.

    fp_radius : :class:`int`, optional
        The radius of the Morgan fingerprint used to calculate
        similarity.

    Returns
    -------
    :class:`float`
        The similarity.

    """

    rdkit_mol1 = mol1.to_rdkit_mol()
    rdkit.SanitizeMol(rdkit_mol1)
    fp1 = rdkit.GetMorganFingerprint(
        mol=rdkit_mol1,
        radius=fp_radius,

    )
    rdkit_mol2 = mol2.to_rdkit_mol()
    rdkit.SanitizeMol(rdkit_mol2)
    fp2 = rdkit.GetMorganFingerprint(
        mol=rdkit_mol2,
        radius=fp_radius,

    )
    return rdkit.DataStructs.DiceSimilarity(fp1, fp2)


def quaternion(u):
    """
    Returns a translation + rotation quaternion.

    Parameters
    ----------
    u : :class:`list` of :class:`float`
        A :class:`list` of length 3 holding the parameter for the
        quarternion.

    References
    ----------
    K. Shoemake, Uniform random rotations, Graphics Gems III,
    pages 124-132. Academic, New York, 1992.

    """

    a, b, c = u
    q = np.zeros(4, np.float64)
    q[0] = np.sqrt(1. - a) * np.sin(2. * np.pi * b)
    q[1] = np.sqrt(1. - a) * np.cos(2. * np.pi * b)
    q[2] = np.sqrt(a) * np.sin(2. * np.pi * c)
    q[3] = np.sqrt(a) * np.cos(2. * np.pi * c)
    return q


def translation_component(q):
    """
    Extracts translation vector from quaternion.

    Parameters
    ----------
    q : :class:`numpy.ndarray`
        A length 4 quaternion.

    Returns
    -------
    :class:`numpy.ndarray`
        The translation vector encoded within `q`.

    """

    rot_epsilon = 1e-6
    q = np.copy(q)
    if q[0] < 0.:
        q = -q
    if q[0] > 1.0:
        q /= np.sqrt(np.dot(q, q))

    theta = 2. * np.arccos(q[0])
    s = np.sqrt(1. - q[0] * q[0])
    if s < rot_epsilon:
        p = 2. * q[1:4]
    else:
        p = q[1:4] / s * theta
    return p


def tar_output():
    """
    Places all the content in the ``output`` folder into a .tgz file.

    Returns
    -------
    None : :class:`NoneType`

    """

    tname = os.path.join('output', 'output.tgz')
    with tarfile.open(tname, 'w:gz') as tar:
        tar.add('output')


@contextmanager
def time_it():
    """
    Times the code executed within the indent.

    This is a context manager so it should be used as:

    .. code-block:: python

        with time_it():
            something1()
            something2()
            something3()

    After all 3 functions are finished and the nested block is exited
    the time taken to process the entire block is printed.

    """

    start = time.time()
    yield
    time_taken = time.time() - start
    m, s = divmod(time_taken, 60)
    h, m = divmod(m, 60)
    print(f'\nTime taken was {int(h)} : {int(m)} : {int(s)}.\n\n')


def vector_angle(vector1, vector2):
    """
    Returns the angle between two vectors in radians.

    Parameters
    ----------
    vector1 : :class:`numpy.ndarray`
        The first vector.

    vector2 : :class:`numpy.ndarray`
        The second vector.

    Returns
    -------
    :class:`float`
        The angle between `vector1` and `vector2` in radians.

    """

    if np.all(np.equal(vector1, vector2)):
        return 0.

    numerator = np.dot(vector1, vector2)
    denominator = np.linalg.norm(vector1) * np.linalg.norm(vector2)
    # This if statement prevents returns of NaN due to floating point
    # inaccuracy.
    term = numerator/denominator
    if term >= 1.:
        return 0.0
    if term <= -1.:
        return np.pi
    return np.arccos(term)


def get_acute_vector(reference, vector):
    if (
        # vector_angle is NaN if reference is [0, 0, 0].
        not np.allclose(reference, [0, 0, 0], atol=1e-5)
        and vector_angle(vector, reference) > np.pi/2
    ):
        return vector * -1
    return vector


def get_plane_normal(points):
    centroid = points.sum(axis=0) / len(points)
    return np.linalg.svd(points - centroid)[-1][2, :]


def cap_absolute_value(value, max_absolute_value=1):
    """
    Returns `value` with absolute value capped at `max_absolute_value`.

    Particularly useful in passing values to trignometric functions
    where numerical errors may result in an argument > 1 being passed
    in.

    This code is modified from the pymatgen source code [1]_.

    Parameters
    ----------
    value : :class:`float`
        Value to cap.

    max_absolute_value : :class:`float`, optional
        Absolute value to cap `value` at.
        Defaults to 1.

    Returns
    -------
    :class:`float`
        `value` capped at `max_absolute_value` with sign preserved.

    References
    ----------
    .. [1] https://pymatgen.org/pymatgen.util.num.html

    """

    return max(min(value, max_absolute_value), -max_absolute_value)



"""
Molecule Mutator
================

.. toctree::
    :maxdepth: 2

    Random Building Block <\
stk.ea.mutation.mutators.molecule.random_building_block\
>
    Random Topology Graph <\
stk.ea.mutation.mutators.molecule.random_topology_graph\
>
    Similar Building Block <\
stk.ea.mutation.mutators.molecule.similar_building_block\
>

"""


class MoleculeMutator:
    """
    Abstract base class for molecule mutators.

    Examples
    --------
    *Subclass Implementation*

    You only need to implement :meth:`.mutate`. The source code of any
    of the classes listed in :mod:`.mutator` can serve as good
    examples.

    """

    def mutate(self, record):
        """
        Return a mutant of `record`.

        Parameters
        ----------
        record : :class:`.MoleculeRecord`
            The molecule to be mutated.

        Returns
        -------
        :class:`.MutationRecord`
            A record of the mutation.

        """

        raise NotImplementedError()



from .topology_graphs import *  # noqa
from .functional_groups import *  # noqa
from .molecules import *  # noqa
from .atoms import *  # noqa
from .bonds import *  # noqa
from .reactions import *  # noqa
from .key_makers import *  # noqa
from .writers import *  # noqa
from .periodic_info import *  # noqa



"""
M12L24
======

"""

from ..cage import Cage
from ..vertices import _NonLinearCageVertex, _LinearCageVertex
from ...topology_graph import Edge


class M12L24(Cage):
    """
    Represents a cage topology graph.

    Metal building blocks with four functional groups are
    required for this topology.

    Ligand building blocks with two functional groups are required for
    this topology.

    When using a :class:`dict` for initialization, a
    :class:`.BuildingBlock` needs to be assigned to each of the
    following numbers:

        | metals: 0 to 11
        | ligands: 12 to 35

    See :class:`.Cage` for more details and examples.

    """

    _vertex_prototypes = (
        _NonLinearCageVertex(0, [1, 0, 0]),
        _NonLinearCageVertex(1, [-1, 0, 0]),
        _NonLinearCageVertex(2, [0, 1, 0]),
        _NonLinearCageVertex(3, [0, -1, 0]),
        _NonLinearCageVertex(4, [0.5, 0.5, 0.707]),
        _NonLinearCageVertex(5, [0.5, -0.5, 0.707]),
        _NonLinearCageVertex(6, [-0.5, 0.5, 0.707]),
        _NonLinearCageVertex(7, [-0.5, -0.5, 0.707]),
        _NonLinearCageVertex(8, [0.5, 0.5, -0.707]),
        _NonLinearCageVertex(9, [0.5, -0.5, -0.707]),
        _NonLinearCageVertex(10, [-0.5, 0.5, -0.707]),
        _NonLinearCageVertex(11, [-0.5, -0.5, -0.707]),

        _LinearCageVertex(12, [0.9, 0.31, 0.31], False),
        _LinearCageVertex(13, [0.9, 0.31, -0.31], False),
        _LinearCageVertex(14, [0.9, -0.31, 0.31], False),
        _LinearCageVertex(15, [0.9, -0.31, -0.31], False),

        _LinearCageVertex(16, [-0.9, 0.31, 0.31], False),
        _LinearCageVertex(17, [-0.9, 0.31, -0.31], False),
        _LinearCageVertex(18, [-0.9, -0.31, 0.31], False),
        _LinearCageVertex(19, [-0.9, -0.31, -0.31], False),

        _LinearCageVertex(20, [0.31, 0.9, 0.31], False),
        _LinearCageVertex(21, [0.31, 0.9, -0.31], False),
        _LinearCageVertex(22, [-0.31, 0.9, 0.31], False),
        _LinearCageVertex(23, [-0.31, 0.9, -0.31], False),

        _LinearCageVertex(24, [0.31, -0.9, 0.31], False),
        _LinearCageVertex(25, [0.31, -0.9, -0.31], False),
        _LinearCageVertex(26, [-0.31, -0.9, 0.31], False),
        _LinearCageVertex(27, [-0.31, -0.9, -0.31], False),

        _LinearCageVertex(28, [0.58, 0, 0.82], False),
        _LinearCageVertex(29, [-0.58, 0, 0.82], False),
        _LinearCageVertex(30, [0, 0.58, 0.82], False),
        _LinearCageVertex(31, [0, -0.58, 0.82], False),
        _LinearCageVertex(32, [0.58, 0, -0.82], False),
        _LinearCageVertex(33, [-0.58, 0, -0.82], False),
        _LinearCageVertex(34, [0, 0.58, -0.82], False),
        _LinearCageVertex(35, [0, -0.58, -0.82], False),
    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[0], _vertex_prototypes[12]),
        Edge(1, _vertex_prototypes[0], _vertex_prototypes[13]),
        Edge(2, _vertex_prototypes[0], _vertex_prototypes[14]),
        Edge(3, _vertex_prototypes[0], _vertex_prototypes[15]),

        Edge(4, _vertex_prototypes[1], _vertex_prototypes[16]),
        Edge(5, _vertex_prototypes[1], _vertex_prototypes[17]),
        Edge(6, _vertex_prototypes[1], _vertex_prototypes[18]),
        Edge(7, _vertex_prototypes[1], _vertex_prototypes[19]),

        Edge(8, _vertex_prototypes[2], _vertex_prototypes[20]),
        Edge(9, _vertex_prototypes[2], _vertex_prototypes[21]),
        Edge(10, _vertex_prototypes[2], _vertex_prototypes[22]),
        Edge(11, _vertex_prototypes[2], _vertex_prototypes[23]),

        Edge(12, _vertex_prototypes[3], _vertex_prototypes[24]),
        Edge(13, _vertex_prototypes[3], _vertex_prototypes[25]),
        Edge(14, _vertex_prototypes[3], _vertex_prototypes[26]),
        Edge(15, _vertex_prototypes[3], _vertex_prototypes[27]),

        Edge(16, _vertex_prototypes[4], _vertex_prototypes[28]),
        Edge(17, _vertex_prototypes[4], _vertex_prototypes[30]),
        Edge(18, _vertex_prototypes[4], _vertex_prototypes[12]),
        Edge(19, _vertex_prototypes[4], _vertex_prototypes[20]),

        Edge(20, _vertex_prototypes[5], _vertex_prototypes[14]),
        Edge(21, _vertex_prototypes[5], _vertex_prototypes[24]),
        Edge(22, _vertex_prototypes[5], _vertex_prototypes[28]),
        Edge(23, _vertex_prototypes[5], _vertex_prototypes[31]),

        Edge(24, _vertex_prototypes[6], _vertex_prototypes[16]),
        Edge(25, _vertex_prototypes[6], _vertex_prototypes[29]),
        Edge(26, _vertex_prototypes[6], _vertex_prototypes[30]),
        Edge(27, _vertex_prototypes[6], _vertex_prototypes[22]),

        Edge(28, _vertex_prototypes[7], _vertex_prototypes[18]),
        Edge(29, _vertex_prototypes[7], _vertex_prototypes[26]),
        Edge(30, _vertex_prototypes[7], _vertex_prototypes[31]),
        Edge(31, _vertex_prototypes[7], _vertex_prototypes[29]),

        Edge(32, _vertex_prototypes[8], _vertex_prototypes[13]),
        Edge(33, _vertex_prototypes[8], _vertex_prototypes[32]),
        Edge(34, _vertex_prototypes[8], _vertex_prototypes[34]),
        Edge(35, _vertex_prototypes[8], _vertex_prototypes[21]),

        Edge(36, _vertex_prototypes[9], _vertex_prototypes[15]),
        Edge(37, _vertex_prototypes[9], _vertex_prototypes[32]),
        Edge(38, _vertex_prototypes[9], _vertex_prototypes[35]),
        Edge(39, _vertex_prototypes[9], _vertex_prototypes[25]),

        Edge(40, _vertex_prototypes[10], _vertex_prototypes[17]),
        Edge(41, _vertex_prototypes[10], _vertex_prototypes[23]),
        Edge(42, _vertex_prototypes[10], _vertex_prototypes[34]),
        Edge(43, _vertex_prototypes[10], _vertex_prototypes[33]),

        Edge(44, _vertex_prototypes[11], _vertex_prototypes[19]),
        Edge(45, _vertex_prototypes[11], _vertex_prototypes[33]),
        Edge(46, _vertex_prototypes[11], _vertex_prototypes[27]),
        Edge(47, _vertex_prototypes[11], _vertex_prototypes[35]),
    )

    _num_windows = 14
    _num_window_types = 2



"""
Value MongoDB
=============

"""

from functools import lru_cache
import warnings

from stk.molecular import InchiKey
from .utilities import HashableDict
from ..value import ValueDatabase


class ValueMongoDb(ValueDatabase):
    """
    Use MongoDB to store and retrieve molecular property values.

    Examples
    --------
    *Storing Molecular Properties in a Database*

    You want to store property values in a database.

    .. code-block:: python

        import stk
        import pymongo

        # Connect to a MongoDB. This example connects to a local
        # MongoDB, but you can connect to a remote DB too with
        # MongoClient() - read the documentation for pymongo to see how
        # to do that.
        client = pymongo.MongoClient()
        db = stk.ValueMongoDb(
            mongo_client=client,
            collection='atom_counts',
        )

        molecule = stk.BuildingBlock('BrCCBr')
        # Add the value to the database.
        db.put(molecule, molecule.get_num_atoms())
        # Retrieve the value from the database.
        num_atoms = db.get(molecule)

        # Works with constructed molecules too.
        polymer = stk.ConstructedMolecule(
            topology_graph=stk.polymer.Linear(
                building_blocks=(
                    stk.BuildingBlock('BrCCBr', [stk.BromoFactory()]),
                ),
                repeating_unit='A',
                num_repeating_units=2',
            ),
        )
        db.put(polymer, polymer.get_num_atoms())
        num_polymer_atoms = db.get(polymer)

    """

    def __init__(
        self,
        mongo_client,
        collection,
        database='stk',
        key_makers=(InchiKey(), ),
        lru_cache_size='',
        put_lru_cache_size=128,
        get_lru_cache_size=128,
        indices=('InChIKey', ),
    ):
        """
        Initialize a :class:`.ValueMongoDb` instance.

        Parameters
        ----------
        mongo_client : :class:`pymongo.MongoClient`
            The database client.

        collection : :class:`str`
            The name of the MongoDB collection used for storing the
            property values.

        database : :class:`str`, optional
            The name of the MongoDB database used for storing the
            property values.

        key_makers : :class:`tuple` of :class:`.MoleculeKeyMaker`
            Used to make the keys of molecules, which the values
            are associated with. If two molecules have the same
            key, they will return the same value from the database.

        lru_cache_size : :class:`int`, optional
            This argument is deprecated and will be removed in any
            version of :mod:`stk` released on, or after, 01/01/21.
            Use the `put_lru_cache_size` and `get_lru_cache_size`
            arguments instead.

            A RAM-based least recently used cache is used to avoid
            reading and writing to the database repeatedly. This sets
            the number of values which fit into the LRU cache. If
            ``None``, the cache size will be unlimited.

        put_lru_cache_size : :class:`int`, optional
            A RAM-based least recently used cache is used to avoid
            writing to the database repeatedly. This sets
            the number of values which fit into the LRU cache. If
            ``None``, the cache size will be unlimited.

        get_lru_cache_size : :class:`int`, optional
            A RAM-based least recently used cache is used to avoid
            reading from the database repeatedly. This sets
            the number of values which fit into the LRU cache. If
            ``None``, the cache size will be unlimited.

        indices : :class:`tuple` of :class:`str`, optional
            The names of molecule keys, on which an index should be
            created, in order to minimize lookup time.

        """

        if lru_cache_size != '':
            warnings.warn(
                'The lru_cache_size argument is deprecated and will '
                'be removed in any version of stk released on, or '
                'after, 01/01/21. Use the put_lru_cache_size and '
                'get_lru_cache_size arguments instead.',
                FutureWarning,
            )
            put_lru_cache_size = lru_cache_size
            get_lru_cache_size = lru_cache_size

        self._values = mongo_client[database][collection]
        self._key_makers = key_makers
        self._put = lru_cache(maxsize=put_lru_cache_size)(self._put)
        self._get = lru_cache(maxsize=get_lru_cache_size)(self._get)

        index_information = self._values.index_information()
        if 'v_1' not in index_information:
            self._values.create_index('v')

        for index in indices:
            # Do not create the same index twice.
            if f'{index}_1' not in index_information:
                self._values.create_index(index)

    def put(self, molecule, value):
        json = {'v': value}
        for key_maker in self._key_makers:
            json[key_maker.get_key_name()] = (
                key_maker.get_key(molecule)
            )
        # lru_cache requires that the parameters to the cached function
        # are hashable objects.
        return self._put(HashableDict(json))

    def _put(self, json):
        keys = dict(json)
        keys.pop('v')

        query = {'$or': []}
        for key, value in keys.items():
            query['$or'].append({key: value})

        self._values.update_many(
            filter=query,
            update={
                '$set': json
            },
            upsert=True,
        )

    def get(self, molecule):

        def make_dict(key_maker):
            return HashableDict({
                key_maker.get_key_name():
                key_maker.get_key(molecule)
            })

        key = {'$or': tuple(map(make_dict, self._key_makers))}
        # lru_cache requires that the parameters to the cached function
        # are hashable objects.
        return self._get(HashableDict(key))

    def _get(self, key):
        value = self._values.find_one(key)
        if value is None:
            raise KeyError(
                'No molecule found in the database with a key of: '
                f'{key}'
            )
        return value['v']



"""
Molecule Key Maker
==================

.. toctree::
    :maxdepth: 2

    InChI <stk.molecular.key_makers.inchi>
    InChIKey <stk.molecular.key_makers.inchi_key>
    SMILES <stk.molecular.key_makers.smiles>

"""


class MoleculeKeyMaker:
    """
    An abstract base class for making :class:`.Molecule` keys.

    Keys are used in :mod:`stk` to determine if two molecules are
    duplicates of each other.

    Notes
    -----
    You might notice that the public methods of this abstract base
    class are implemented. This is purely for convenience when
    implementing subclasses. The implemented public methods are
    simply default implementations, which can be safely ignored or
    overridden, when implementing subclasses.

    Examples
    --------
    *Subclass Implementation*

    The source code of any of the subclasses, listed in
    :mod:`molecule_key_maker <.key_makers.molecule>`, can
    serve as good examples.

    *Creating a New Key Maker Directly*

    Apart from using the subclasses provided, a
    :class:`.MoleculeKeyMaker` can be used directly , if you don't
    feel like writing a subclass

    .. code-block:: python

        import stk

        # Create a MoleculeKeyMaker instance with a custom get_key
        # method.
        get_num_atoms = stk.MoleculeKeyMaker(
            key_name='num_atoms',
            get_key=lambda molecule: molecule.get_num_atoms(),
        )

        # Use the MoleculeKeyMaker instance.
        jsonizer = stk.MoleculeJsonizer(
            key_makers=(get_num_atoms, ),
        )
        # Get the JSON representation of a molecule.
        json = jsonizer.to_json(stk.BuildingBlock('NCCN'))

    """

    def __init__(self, key_name, get_key):
        """
        Initialize a :class:`.MoleculeKeyMaker` instance.

        Parameters
        ----------
        key_name : :class:`str`
            The name of the key.

        get_key : :class:`callable`
            Takes a single parameter, `molecule`, and returns the
            key to use for that molecule. The value passed to the
            parameter must be a :class:`.Molecule` instance.

        """

        self._key_name = key_name
        self._get_key = get_key

    def get_key_name(self):
        """
        Get the name of the key.

        Returns
        -------
        :class:`str`
            The name of the key.

        """

        return self._key_name

    def get_key(self, molecule):
        """
        Get the key of `molecule`.

        Parameters
        ----------
        molecule : :class:`.Molecule`
            The molecule for which a key is needed.

        Returns
        -------
        :class:`object`
            The key of `molecule`.

        """

        return self._get_key(molecule)

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return (
            f'{self.__class__.__name__}('
            f'{self._key_name!r}, {self._get_key!r}'
            ')'
        )



"""
Six Plus Eight
==============

"""

from ..cage import Cage
from ..vertices import _NonLinearCageVertex
from ...topology_graph import Edge


class SixPlusEight(Cage):
    """
    Represents a cage topology graph.

    Building blocks with three and four functional groups are required
    for this topology graph.

    See :class:`.Cage` for more details and examples.

    """

    _x = 1
    _vertex_prototypes = (
        _NonLinearCageVertex(0, [-_x, _x, 0]),
        _NonLinearCageVertex(1, [-_x, -_x, 0]),
        _NonLinearCageVertex(2, [_x, _x, 0]),
        _NonLinearCageVertex(3, [_x, -_x, 0]),

        _NonLinearCageVertex(4, [0, 0, _x]),
        _NonLinearCageVertex(5, [0, 0, -_x]),
    )

    _vertex_prototypes = (
        *_vertex_prototypes,
        _NonLinearCageVertex.init_at_center(
            id=6,
            vertices=(
                _vertex_prototypes[0],
                _vertex_prototypes[4],
                _vertex_prototypes[1]
            ),

        ),
        _NonLinearCageVertex.init_at_center(
            id=7,
            vertices=(
                _vertex_prototypes[1],
                _vertex_prototypes[4],
                _vertex_prototypes[3]
            ),
        ),
        _NonLinearCageVertex.init_at_center(
            id=8,
            vertices=(
                _vertex_prototypes[4],
                _vertex_prototypes[3],
                _vertex_prototypes[2]
            ),
        ),
        _NonLinearCageVertex.init_at_center(
            id=9,
            vertices=(
                _vertex_prototypes[4],
                _vertex_prototypes[2],
                _vertex_prototypes[0]
            ),
        ),

        _NonLinearCageVertex.init_at_center(
            id=10,
            vertices=(
                _vertex_prototypes[0],
                _vertex_prototypes[5],
                _vertex_prototypes[1]
            ),
        ),
        _NonLinearCageVertex.init_at_center(
            id=11,
            vertices=(
                _vertex_prototypes[5],
                _vertex_prototypes[1],
                _vertex_prototypes[3]
            ),
        ),
        _NonLinearCageVertex.init_at_center(
            id=12,
            vertices=(
                _vertex_prototypes[3],
                _vertex_prototypes[5],
                _vertex_prototypes[2]
            ),
        ),
        _NonLinearCageVertex.init_at_center(
            id=13,
            vertices=(
                _vertex_prototypes[2],
                _vertex_prototypes[5],
                _vertex_prototypes[0]
            ),
        ),
    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[6], _vertex_prototypes[0]),
        Edge(1, _vertex_prototypes[6], _vertex_prototypes[4]),
        Edge(2, _vertex_prototypes[6], _vertex_prototypes[1]),

        Edge(3, _vertex_prototypes[7], _vertex_prototypes[1]),
        Edge(4, _vertex_prototypes[7], _vertex_prototypes[4]),
        Edge(5, _vertex_prototypes[7], _vertex_prototypes[3]),

        Edge(6, _vertex_prototypes[8], _vertex_prototypes[4]),
        Edge(7, _vertex_prototypes[8], _vertex_prototypes[3]),
        Edge(8, _vertex_prototypes[8], _vertex_prototypes[2]),

        Edge(9, _vertex_prototypes[9], _vertex_prototypes[4]),
        Edge(10, _vertex_prototypes[9], _vertex_prototypes[2]),
        Edge(11, _vertex_prototypes[9], _vertex_prototypes[0]),

        Edge(12, _vertex_prototypes[10], _vertex_prototypes[0]),
        Edge(13, _vertex_prototypes[10], _vertex_prototypes[5]),
        Edge(14, _vertex_prototypes[10], _vertex_prototypes[1]),

        Edge(15, _vertex_prototypes[11], _vertex_prototypes[5]),
        Edge(16, _vertex_prototypes[11], _vertex_prototypes[1]),
        Edge(17, _vertex_prototypes[11], _vertex_prototypes[3]),

        Edge(18, _vertex_prototypes[12], _vertex_prototypes[3]),
        Edge(19, _vertex_prototypes[12], _vertex_prototypes[5]),
        Edge(20, _vertex_prototypes[12], _vertex_prototypes[2]),

        Edge(21, _vertex_prototypes[13], _vertex_prototypes[2]),
        Edge(22, _vertex_prototypes[13], _vertex_prototypes[5]),
        Edge(23, _vertex_prototypes[13], _vertex_prototypes[0]),

    )

    _num_windows = 12
    _num_window_types = 1



from .crossover import *  # noqa
from .fitness_calculators import *  # noqa
from .mutation import *  # noqa
from .fitness_normalizers import *  # noqa
from .plotters import *  # noqa
from .selection import *  # noqa
from .generation import *  # noqa
from .evolutionary_algorithm import *  # noqa
from .molecule_records import *  # noqa



"""
M4L4 Tetrahedron
================

"""

import numpy as np

from ..cage import Cage
from ..vertices import _NonLinearCageVertex
from ...topology_graph import Edge


class M4L4Tetrahedron(Cage):
    """
    Represents a cage topology graph.

    Building blocks with three functional groups are required for this
    topology.

    See :class:`.Cage` for more details and examples.

    """

    _vertex_prototypes = (
        _NonLinearCageVertex(0, [0, 0, np.sqrt(6)/2]),
        _NonLinearCageVertex(1, [-1, -np.sqrt(3)/3, -np.sqrt(6)/6]),
        _NonLinearCageVertex(2, [1, -np.sqrt(3)/3, -np.sqrt(6)/6]),
        _NonLinearCageVertex(3, [0, 2*np.sqrt(3)/3, -np.sqrt(6)/6]),
    )

    _vertex_prototypes = (
        *_vertex_prototypes,

        _NonLinearCageVertex.init_at_center(
            id=4,
            vertices=(
                _vertex_prototypes[0],
                _vertex_prototypes[1],
                _vertex_prototypes[2],
            ),
        ),
        _NonLinearCageVertex.init_at_center(
            id=5,
            vertices=(
                _vertex_prototypes[0],
                _vertex_prototypes[1],
                _vertex_prototypes[3],
            ),
        ),
        _NonLinearCageVertex.init_at_center(
            id=6,
            vertices=(
                _vertex_prototypes[0],
                _vertex_prototypes[2],
                _vertex_prototypes[3],
            ),
        ),
        _NonLinearCageVertex.init_at_center(
            id=7,
            vertices=(
                _vertex_prototypes[1],
                _vertex_prototypes[2],
                _vertex_prototypes[3],
            ),
        ),
    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[0], _vertex_prototypes[4]),
        Edge(1, _vertex_prototypes[0], _vertex_prototypes[5]),
        Edge(2, _vertex_prototypes[0], _vertex_prototypes[6]),
        Edge(3, _vertex_prototypes[1], _vertex_prototypes[4]),
        Edge(4, _vertex_prototypes[1], _vertex_prototypes[5]),
        Edge(5, _vertex_prototypes[1], _vertex_prototypes[7]),
        Edge(6, _vertex_prototypes[2], _vertex_prototypes[4]),
        Edge(7, _vertex_prototypes[2], _vertex_prototypes[6]),
        Edge(8, _vertex_prototypes[2], _vertex_prototypes[7]),
        Edge(9, _vertex_prototypes[3], _vertex_prototypes[5]),
        Edge(10, _vertex_prototypes[3], _vertex_prototypes[6]),
        Edge(11, _vertex_prototypes[3], _vertex_prototypes[7]),
    )

    _num_windows = 4
    _num_window_types = 1



"""
Constructed Molecule JSONizer
=============================

"""

from stk.molecular import (
    MoleculeKeyMaker,
    InchiKey,
)
from ..molecule import MoleculeJsonizer


class ConstructedMoleculeJsonizer:
    """
    Abstract base class for creating JSONs of constructed molecules.

    See Also
    --------
    :class:`.MoleculeJsonizer`

    Notes
    -----
    You might notice that the public methods of this abstract base
    class are implemented. These are just default implementations,
    which can be safely ignored or overridden, when implementing
    subclasses. However, the default implementation can be used
    directly, if it suits your needs.

    Examples
    --------
    You want get a JSON representation of a
    :class:`.ConstructedMolecule`

    .. code-block:: python

        import stk

        # Make the molecule you want jsonize.

        polymer_topology_graph = stk.polymer.Linear(
            building_blocks=(
                stk.BuildingBlock('BrCCBr', [stk.BromoFactory()],),
            ),
            repeating_unit='A',
            num_repeating_units=3,
        )
        polymer = stk.ConstructedMolecule(polymer_topology_graph)


        # Make a JSONizer.
        jsonizer = stk.ConstructedMoleculeJsonizer()
        # Get the JSON.
        json = jsonizer.to_json(polymer)

    """

    def __init__(
        self,
        key_makers=(InchiKey(), ),
    ):
        """
        Initializes a :class:`.ConstructedMoleculeJsonizer`.

        Parameters
        ----------
        key_makers : :class:`tuple` of :class:`.MoleculeKeyMaker`
            Used to make the keys of molecules, which should be
            included in their JSON representations. Keys allow
            molecular data to reference itself when split across
            multiple JSONs.

        """

        self._jsonizer = MoleculeJsonizer(key_makers=())
        self._key_makers = key_makers

    def to_json(self, molecule):
        """
        Serialize `molecule`.

        Parameters
        ----------
        molecule : :class:`.ConstructedMolecule`
            The constructed molecule to serialize.

        Returns
        -------
        :class:`dict`
            A JSON representation of `molecule`.

        """

        def get_keys(building_block):
            return {
                key_maker.get_key_name():
                    key_maker.get_key(building_block)
                for key_maker in self._key_makers
                if isinstance(key_maker, MoleculeKeyMaker)
            }

        building_block_indices = {
            building_block: index
            for index, building_block
            in enumerate(molecule.get_building_blocks())
        }
        building_block_indices[None] = None

        def atom_info_to_json(atom_info):
            return (
                building_block_indices[atom_info.get_building_block()],
                atom_info.get_building_block_id(),
            )

        def bond_info_to_json(bond_info):
            return (
                building_block_indices[bond_info.get_building_block()],
                bond_info.get_building_block_id(),
            )

        molecule_json = self._jsonizer.to_json(molecule)
        constructed_molecule_json = {
            'BB': tuple(map(
                get_keys,
                molecule.get_building_blocks(),
            )),
            'aI': tuple(map(
                atom_info_to_json,
                molecule.get_atom_infos(),
            )),
            'bI': tuple(map(
                bond_info_to_json,
                molecule.get_bond_infos(),
            )),
            'nBB': tuple(map(
                molecule.get_num_building_block,
                molecule.get_building_blocks(),
            )),
        }
        for key_maker in self._key_makers:
            key_name = key_maker.get_key_name()
            key = key_maker.get_key(molecule)
            molecule_json['molecule'][key_name] = key
            molecule_json['matrix'][key_name] = key
            constructed_molecule_json[key_name] = key

        building_block_jsons = tuple(map(
            self._jsonizer.to_json,
            molecule.get_building_blocks(),
        ))

        def is_molecule_key_maker(key_maker):
            return isinstance(key_maker, MoleculeKeyMaker)

        for key_maker in filter(
            is_molecule_key_maker,
            self._key_makers,
        ):
            key_name = key_maker.get_key_name()
            for building_block, json in zip(
                molecule.get_building_blocks(),
                building_block_jsons,
            ):
                key = key_maker.get_key(building_block)
                json['molecule'][key_name] = key
                json['matrix'][key_name] = key

        return {
            'molecule': molecule_json['molecule'],
            'constructedMolecule': constructed_molecule_json,
            'matrix': molecule_json['matrix'],
            'buildingBlocks': building_block_jsons,
        }

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return f'{self.__class__.__name__}({self._key_makers!r})'



"""
Constructed Molecule DeJSONizer
===============================

"""

import numpy as np
from stk.molecular import ConstructedMolecule

from .utilities import to_atom, to_bond, to_atom_info, to_bond_info
from .molecule import MoleculeDejsonizer


class ConstructedMoleculeDejsonizer:
    """
    Abstract base class for creating constructed molecules from JSONs.

    See Also
    --------
    :class:`.MoleculeDejsonizer`

    Notes
    -----
    You might notice that the public methods of this abstract base
    class are implemented. These are just default implementations,
    which can be safely ignored or overridden, when implementing
    subclasses. However, the default implementation can be used
    directly, if it suits your needs.

    """

    def __init__(self):
        """
        Initialize a :class:`.ConstructedMoleculeDejsonizer` instance.

        """

        self._dejsonizer = MoleculeDejsonizer()

    def from_json(self, json):
        """
        Get a :class:`.ConstructedMolecule` from a JSON.

        Parameters
        ----------
        json : :class:`dict`
            A JSON of the constructed molecule.

        Returns
        -------
        :class:`.ConstructedMolecule`
            The constructed molecule.

        """

        building_blocks = tuple(map(
            self._dejsonizer.from_json,
            json['buildingBlocks'],
        ))

        num_building_blocks = (
            (building_block, num)
            for building_block, num in zip(
                building_blocks,
                json['constructedMolecule']['nBB'],
            )
        )
        atoms = tuple(
            to_atom(atom_id, atom_json)
            for atom_id, atom_json in enumerate(json['molecule']['a'])
        )
        bonds = tuple(
            to_bond(atoms, bond_json)
            for bond_json in json['molecule']['b']
        )
        return ConstructedMolecule.init(
            atoms=atoms,
            bonds=bonds,
            position_matrix=np.array(json['matrix']['m']),
            atom_infos=tuple(
                to_atom_info(
                    building_blocks=building_blocks,
                    atom=atoms[atom_id],
                    json=atom_info_json,
                )
                for atom_id, atom_info_json
                in enumerate(json['constructedMolecule']['aI'])
            ),
            bond_infos=tuple(
                to_bond_info(
                    building_blocks=building_blocks,
                    bond=bonds[bond_id],
                    json=bond_info_json,
                )
                for bond_id, bond_info_json
                in enumerate(json['constructedMolecule']['bI'])
            ),
            num_building_blocks=num_building_blocks,
        )

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return f'{self.__class__.__name__}()'



"""
Two Plus Two
============

"""

import numpy as np

from ..cage import Cage
from ..vertices import _NonLinearCageVertex
from ...topology_graph import Edge


class TwoPlusTwo(Cage):
    """
    Represents a tetrahedron cage topology graph.

    See :class:`.Cage` for more details and examples.

    """

    _x = 1
    _vertex_prototypes = (
        _NonLinearCageVertex(0, [_x, 0, -_x/np.sqrt(2)], False),
        _NonLinearCageVertex(1, [-_x, 0, -_x/np.sqrt(2)], False),
        _NonLinearCageVertex(2, [0, _x, _x/np.sqrt(2)], False),
        _NonLinearCageVertex(3, [0, -_x, _x/np.sqrt(2)], False)
    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[0], _vertex_prototypes[1]),
        Edge(1, _vertex_prototypes[0], _vertex_prototypes[2]),
        Edge(2, _vertex_prototypes[0], _vertex_prototypes[3]),

        Edge(3, _vertex_prototypes[1], _vertex_prototypes[2]),
        Edge(4, _vertex_prototypes[1], _vertex_prototypes[3]),

        Edge(5, _vertex_prototypes[2], _vertex_prototypes[3])
    )

    _num_windows = 4
    _num_window_types = 1



"""
Random Mutator
==============

"""


import numpy as np


class RandomMutator:
    """
    Use some other mutator at random.

    Examples
    --------
    *Use One of Several Mutators at Random*

    .. code-block:: python

        import stk

        mutator = stk.RandomMutator(
            mutators=(
                stk.RandomBuildingBlock(...),
                stk.SimilarBuildingBlock(...),
                stk.RandomBuildingBlock(...),
            ),
        )
        # Use one of the component mutators at random.
        mutation_record = mutator.mutate(record)
        # A different mutator may get selected at random the second,
        # third, etc, time.
        mutation_record2 = mutator.mutate(record)

    """

    def __init__(self, mutators, weights=None, random_seed=None):
        """
        Initialize a :class:`.RandomMutator` instance.

        Parameters
        ----------
        mutators : :class:`tuple`
            Holds instances which have :meth:`.mutate` method. The
            :meth:`.mutate` method must return an instance of
            :class:`.MutationRecord`.

        weights : :class:`tuple` of :class:`float`, optional
            For each mutator, the probability that it will be chosen
            whenever :meth:`.mutate` is called.
            If ``None`` all `mutators` will have equal chance of being
            selected.

        random_seed : :class:`int`, optional
            The random seed to use.

        """

        self._mutators = mutators
        self._weights = weights
        self._generator = np.random.RandomState(random_seed)

    def mutate(self, record):
        """
        Return a mutant of `record`.

        Parameters
        ----------
        record : :class:`.MoleculeRecord`
            The molecule to be mutated.

        Returns
        -------
        :class:`.MutationRecord`
            A record of the mutation. The exact subclass of
            :class:`.MutationRecord` depends on which mutator was
            used.

        """

        mutator = self._generator.choice(
            a=self._mutators,
            p=self._weights,
        )
        return mutator.mutate(record)



"""
Molecule Crosser
================

.. toctree::
    :maxdepth: 2

    Genetic Recombination <\
stk.ea.crossover.crossers.molecule.genetic_recombination\
>

"""


class MoleculeCrosser:
    """
    Abstract base class for molecule crossers.

    Crossers take multiple molecules and recombine them to make
    new, offspring, molecules.

    Examples
    --------
    *Subclass Implementation*

    You only need to implement :meth:`.cross`. The source code of any
    of the classes listed in :mod:`.crosser` can serve as good
    examples.

    """

    def cross(self, records):
        """
        Cross `records`.

        Parameters
        ----------
        records : :class:`iterable` of :class:`.MoleculeRecord`
            The molecule records on which a crossover operation is
            performed.

        Yields
        -------
        :class:`.CrossoverRecord`
            A record of a crossover operation.

        """

        raise NotImplementedError()



"""
Random Crosser
==============

"""


import numpy as np


class RandomCrosser:
    """
    Use some other crosser at random.

    Examples
    --------
    *Use One of Several Crossers at Random*

    .. code-block:: python

        import stk

        crosser = stk.RandomCrosser(
            crossers=(
                # Assume each crosser crosses in a different way.
                stk.GeneticRecombination(...),
                stk.GeneticRecombination(...),
            ),
        )
        # Use one of the component crossers at random.
        crossover_record1 = crosser.cross(record)
        # A different crosser may get selected at random the second,
        # third, etc, time.
        crossover_record2 = crosser.cross(record)

    """

    def __init__(self, crossers, weights=None, random_seed=None):
        """
        Initialize a :class:`.RandomMutator` instance.

        Parameters
        ----------
        crossers : :class:`tuple`
            Holds instances which have :meth:`.cross` method. The
            :meth:`.cross` method must return an instance of
            :class:`.CrossoverRecord`.

        weights : :class:`tuple` of :class:`float`, optional
            For each mutator, the probability that it will be chosen
            whenever :meth:`.mutate` is called.
            If ``None`` all `mutators` will have equal chance of being
            selected.

        random_seed : :class:`int`, optional
            The random seed to use.

        """

        self._crossers = crossers
        self._weights = weights
        self._generator = np.random.RandomState(random_seed)

    def cross(self, records):
        """
        Cross `records`.

        Parameters
        ----------
        records : :class:`iterable` of :class:`.MoleculeRecord`
            The molecule records on which a crossover operation is
            performed.

        Yields
        -------
        :class:`.CrossoverRecord`
            A record of a crossover operation.

        """

        crosser = self._generator.choice(
            a=self._crossers,
            p=self._weights,
        )
        return crosser.cross(records)



"""
Molecule Mutation Record
========================

"""


class MutationRecord:
    """
    Abstract base class for a record of a mutation operation.

    Notes
    -----
    You might notice that the public methods of this abstract base
    class are implemented. This is just a default implementation, which
    can be used directly by users and subclasses, but can also be
    freely replaced during subclass implementation, if need be.

    """

    def __init__(self, molecule_record, mutator_name):
        """
        Initialize a :class:`.MutationRecord` instance.

        Parameters
        ----------
        molecule_record : :class:`.MoleculeRecord`
            The molecule produced by the mutation operation.

        mutator_name : :class:`str`
            The name of the mutator which carried out the mutator.

        """

        self._molecule_record = molecule_record
        self._mutator_name = mutator_name

    def get_molecule_record(self):
        """
        Get the :class:`.MoleculeRecord` produced by the mutation.

        Returns
        -------
        :class:`.MoleculeRecord`
            The molecule record.

        """

        return self._molecule_record

    def get_mutator_name(self):
        """
        Get the name of the mutator which carried out the mutation.

        Returns
        -------
        :class:`str`
            The name of the mutator.

        """

        return self._mutator_name



"""
M2L4 Lantern
============

"""

from ..cage import Cage
from ..vertices import _NonLinearCageVertex, _LinearCageVertex
from ...topology_graph import Edge


class M2L4Lantern(Cage):
    """
    Represents a cage topology graph.

    Metal building blocks with four functional groups are
    required for this topology.

    Ligand building blocks with two functional groups are required for
    this topology.

    When using a :class:`dict` for initialization, a
    :class:`.BuildingBlock` needs to be assigned to each of the
    following numbers:

        | metals: (0, 1)
        | ligands: (2, 3, 4, 5)

    See :class:`.Cage` for more details and examples.

    """

    _vertex_prototypes = (
        _NonLinearCageVertex(0, [0, 0.5, 0]),
        _NonLinearCageVertex(1, [0, -0.5, 0]),

        _LinearCageVertex(2, [1, 0, 0], False),
        _LinearCageVertex(3, [0, 0, 1], False),
        _LinearCageVertex(4, [-1, 0, 0], False),
        _LinearCageVertex(5, [0, 0, -1], False),
    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[0], _vertex_prototypes[2]),
        Edge(1, _vertex_prototypes[0], _vertex_prototypes[3]),
        Edge(2, _vertex_prototypes[0], _vertex_prototypes[4]),
        Edge(3, _vertex_prototypes[0], _vertex_prototypes[5]),

        Edge(4, _vertex_prototypes[1], _vertex_prototypes[2]),
        Edge(5, _vertex_prototypes[1], _vertex_prototypes[3]),
        Edge(6, _vertex_prototypes[1], _vertex_prototypes[4]),
        Edge(7, _vertex_prototypes[1], _vertex_prototypes[5]),
    )

    _num_windows = 4
    _num_window_types = 1



"""
Porphyrin
=========

"""

from ..metal_complex import MetalComplex
from ..vertices import _MetalVertex
from ...topology_graph import Edge


class Porphyrin(MetalComplex):
    """
    Represents a metal complex topology graph.

    Metal building blocks with at least four functional groups are
    required for this topology.

    Ligand building blocks with two functional groups are required for
    this topology graph.

    When using a :class:`dict` for initialization, a
    :class:`.BuildingBlock` needs to be assigned to each of the
    following numbers:

        | metals: (0, )
        | ligands: (0, )

    See :class:`.MetalComplex` for more details and examples.

    """

    _metal_vertex_prototypes = (_MetalVertex(0, [0, 0, 0]), )
    _ligand_vertex_prototypes = (_MetalVertex(1, [0, 0, 0]), )

    _edge_prototypes = (
        Edge(
            id=0,
            vertex1=_metal_vertex_prototypes[0],
            vertex2=_ligand_vertex_prototypes[0],
            position=[0.1, 0, 0],
        ),
        Edge(
            id=1,
            vertex1=_metal_vertex_prototypes[0],
            vertex2=_ligand_vertex_prototypes[0],
            position=[0, 0.1, 0],
        ),
        Edge(
            id=2,
            vertex1=_metal_vertex_prototypes[0],
            vertex2=_ligand_vertex_prototypes[0],
            position=[-0.1, 0, 0],
        ),
        Edge(
            id=3,
            vertex1=_metal_vertex_prototypes[0],
            vertex2=_ligand_vertex_prototypes[0],
            position=[0, -0.1, 0],
        ),
    )



"""
InChIKey
========

"""

from .molecule import MoleculeKeyMaker
from .utilities import get_inchi_key


class InchiKey(MoleculeKeyMaker):
    """
    Used to get the InChIKey of molecules.

    Examples
    --------
    *Adding InChIKey to a Molecule's JSON*

    You want to use the InChIKey as part of a JSON representation of a
    molecule

    .. code-block:: python

        import stk

        jsonizer = stk.MoleculeJsonizer(
            key_makers=(stk.InchiKey(), ),
        )
        # Get the JSON representation, including an InChIKey.
        json = jsonizer.to_json(stk.BuildingBlock('NCCN'))

    """

    def __init__(self):
        """
        Initialize an :class:`InchiKey` instance.

        """

        super().__init__('InChIKey', get_inchi_key)

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return 'InchiKey()'



"""
Four Plus Four
==============

"""

from ..cage import Cage
from ..vertices import _NonLinearCageVertex
from ...topology_graph import Edge


class FourPlusFour(Cage):
    """
    Represents a cube cage topology graph.

    See :class:`.Cage` for more details and examples.

    """

    _x = 1
    _vertex_prototypes = (
        _NonLinearCageVertex(0, [-_x, _x, -_x], False),
        _NonLinearCageVertex(1, [-_x, -_x, -_x], False),
        _NonLinearCageVertex(2, [_x, _x, -_x], False),
        _NonLinearCageVertex(3, [_x, -_x, -_x], False),

        _NonLinearCageVertex(4, [-_x, _x, _x], False),
        _NonLinearCageVertex(5, [-_x, -_x, _x], False),
        _NonLinearCageVertex(6, [_x, _x, _x], False),
        _NonLinearCageVertex(7, [_x, -_x, _x], False)
    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[0], _vertex_prototypes[1]),
        Edge(1, _vertex_prototypes[0], _vertex_prototypes[2]),
        Edge(2, _vertex_prototypes[0], _vertex_prototypes[4]),
        Edge(3, _vertex_prototypes[1], _vertex_prototypes[3]),
        Edge(4, _vertex_prototypes[1], _vertex_prototypes[5]),
        Edge(5, _vertex_prototypes[2], _vertex_prototypes[6]),
        Edge(6, _vertex_prototypes[2], _vertex_prototypes[3]),
        Edge(7, _vertex_prototypes[3], _vertex_prototypes[7]),
        Edge(8, _vertex_prototypes[4], _vertex_prototypes[6]),
        Edge(9, _vertex_prototypes[4], _vertex_prototypes[5]),
        Edge(10, _vertex_prototypes[5], _vertex_prototypes[7]),
        Edge(11, _vertex_prototypes[6], _vertex_prototypes[7])
    )

    _num_windows = 6
    _num_window_types = 1



"""
Atom
====

"""


class Atom:
    """
    An abstract base class for atoms.

    A subclass is made for each element. The name of each subclass is
    the periodic table symbol of that element.

    Atoms of a particular element can be made with this
    class or with the subclass representing that element.

    Examples
    --------
    *Initialization.*

    Initialization of an :class:`.Atom` can happen in one of two ways.
    The atom can be initialized through the :class:`.Atom` class or
    through the class representing the element.

    .. code-block:: python

        import stk

        # h0 is an instance of the H class.
        h0 = stk.Atom(id=0, atomic_number=1)

        # h1 is also an instance of the H class.
        h1 = stk.H(id=1)

    When the class correspnding to the element is used directly, the
    ``atomic_number`` is not provided. Here are a few more examples.

    .. code-block:: python

        # Both he0 and he1 are instances of the He class.
        he0 = stk.Atom(id=2, atomic_number=2)
        he1 = stk.He(id=3)

        # Both c0 and c1 are instances of the
        # C class.
        c0 = stk.Atom(id=4, atomic_number=6)
        c1 = stk.C(id=5)

    """

    # Maps each atomic number (int) to the relevant Atom subclass.
    _elements = {}

    def __init_subclass__(cls, **kwargs):
        # Replace the default __init__() method of the subclass with
        # _subclass_init(). This is because the default __init__()
        # method takes an atomic_number parameter, but
        # _subclass_init() does not.
        cls.__init__ = cls._subclass_init
        cls._elements[cls._atomic_number] = cls

    @staticmethod
    def _subclass_init(self, id, charge=0):
        """
        Initialize an atom of the element.

        Parameters
        ----------
        id : :class:`int`
            The id of the atom.

        charge : :class:`int`
            The formal charge.

        """

        Atom.__init__(self, id, self._atomic_number, charge)

    def __init__(self, id, atomic_number, charge=0):
        """
        Initialize an :class:`Atom`.

        Parameters
        ----------
        id : :class:`int`
            The id of the atom.

        atomic_number : :class:`int`
            The atomic number.

        charge : :class:`int`
            The formal charge.

        """

        self.__class__ = self._elements[atomic_number]
        self._id = id
        self._charge = charge

    def get_id(self):
        """
        Get the id of the atom.

        Returns
        -------
        :class:`int`
            The id.

        """

        return self._id

    def _with_id(self, id):
        """
        Modify the atom.

        """

        self._id = id
        return self

    def with_id(self, id):
        """
        Get a clone but with a different id.

        Returns
        -------
        :class:`.Atom`
            A clone with a new id. Has the same type as the original
            atom.

        """

        return self.clone()._with_id(id)

    def get_atomic_number(self):
        """
        Get the atomic number of the atom.

        Returns
        -------
        :class:`int`
            The atomic number.

        """

        return self._atomic_number

    def get_charge(self):
        """
        Get the charge of the atom.

        Returns
        -------
        :class:`int`
            The charge.

        """

        return self._charge

    def __repr__(self):
        charge = (
            f', charge={self._charge}' if self._charge != 0 else ''
        )
        return f'{self.__class__.__name__}({self._id}{charge})'

    def __str__(self):
        return repr(self)

    def clone(self):
        """
        Return a clone.

        Returns
        -------
        :class:`.Atom`
            The clone. It has the same type as the original atom.

        """

        clone = self.__class__.__new__(self.__class__)
        clone._id = self._id
        clone._charge = self._charge
        return clone



from setuptools import find_packages, setup
import re
from os.path import join


def get_version():
    with open(join('src', 'stk', '__init__.py'), 'r') as f:
        content = f.read()
    p = re.compile(r'^__version__ = [\'"]([^\'\"]*)[\'"]', re.M)
    return p.search(content).group(1)


setup(
    name='stk',
    author='Lukas Turcani',
    author_email='lukasturcani93@gmail.com',
    url='https://www.github.com/lukasturcani/stk',
    version=get_version(),
    package_dir={'': 'src'},
    packages=find_packages(where='src'),
    install_requires=(
        'scipy',
        'matplotlib',
        'pandas',
        'pathos',
        'seaborn',
        'numpy',
        'pymongo',
        'pymongo[srv]',
    ),
    python_requires='>=3.7',
)



"""
Primary Amino Factory
=====================

"""

from .functional_group_factory import FunctionalGroupFactory
from .utilities import _get_atom_ids
from ..functional_groups import PrimaryAmino


class PrimaryAminoFactory(FunctionalGroupFactory):
    """
    Creates :class:`.PrimaryAmino` instances.

    Creates functional groups from substructures, which match the
    ``[*][N]([H])[H]`` functional group string.

    Examples
    --------
    You want to create a building block which has
    :class:`.PrimaryAmino` functional groups. You want the nitrogen
    atom to be the *bonder* atom, and the hydrogen atoms to be the
    *deleter* atoms.

    .. code-block:: python

        import stk

        building_block = stk.BuildingBlock(
            smiles='NCCCCN',
            functional_groups=(stk.PrimaryAminoFactory(), ),
        )

    You want to create a building block which has
    :class:`.PrimaryAmino` functional groups. You want the non-hydrogen
    atom bonded to nitrogen to be the *bonder* atom and the
    nitrogen and hydrogen atoms to be *deleter* atoms.

    .. code-block:: python

        import stk

        primary_amino_factory = stk.PrimaryAminoFactory(
            # The index of the atom attached to the nitrogen is 0 in
            # the functional group string (see docstring).
            bonders=(0, ),
            # The indices of the nitrogen and hydrogen atoms in the
            # functional group string (see docstring) are 1, 2 and 3.
            deleters=(1, 2, 3),
        )
        building_block = stk.BuildingBlock(
            smiles='NCCCCN',
            functional_groups=(primary_amino_factory, ),
        )


    See Also
    --------
    :class:`.GenericFunctionalGroup`
        Defines *bonders* and  *deleters*.

    """

    def __init__(self, bonders=(1, ), deleters=(2, 3), placers=None):
        """
        Initialize a :class:`.PrimaryAminoFactory` instance.

        Parameters
        ----------
        bonders : :class:`tuple` of :class:`int`
            The indices of atoms in the functional group string, which
            are *bonder* atoms.

        deleters : :class:`tuple` of :class:`int`
            The indices of atoms in the functional group string, which
            are *deleter* atoms.

        placers : :class:`tuple` of :class:`int`, optional
            The indices of atoms in the functional group string, which
            are *placer* atoms. If ``None``, `bonders` will be used.

        """

        """
        Initialize an :class:`.AmineFactory`.

        """

        self._bonders = bonders
        self._deleters = deleters
        self._placers = bonders if placers is None else placers

    def get_functional_groups(self, molecule):
        for atom_ids in _get_atom_ids('[*][N]([H])[H]', molecule):
            atoms = tuple(molecule.get_atoms(atom_ids))
            yield PrimaryAmino(
                nitrogen=atoms[1],
                hydrogen1=atoms[2],
                hydrogen2=atoms[3],
                atom=atoms[0],
                bonders=tuple(atoms[i] for i in self._bonders),
                deleters=tuple(atoms[i] for i in self._deleters),
                placers=tuple(atoms[i] for i in self._placers),
            )



"""
M6L12 Cube
==========

"""

import numpy as np

from ..cage import Cage
from ..vertices import _NonLinearCageVertex, _LinearCageVertex
from ...topology_graph import Edge


class M6L12Cube(Cage):
    """
    Represents a cage topology graph.

    Metal building blocks with four functional groups are
    required for this topology.

    Ligand building blocks with two functional groups are required for
    this topology.

    When using a :class:`dict` for initialization, a
    :class:`.BuildingBlock` needs to be assigned to each of the
    following numbers:

        | metals: (0, 1, 2, 3, 4, 5)
        | ligands: (6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17)

    See :class:`.Cage` for more details and examples.

    """

    _x = np.sqrt(2)
    _vertex_prototypes = (
        _NonLinearCageVertex(0, [_x, 0, 0]),
        _NonLinearCageVertex(1, [0, _x, 0]),
        _NonLinearCageVertex(2, [-_x, 0, 0]),
        _NonLinearCageVertex(3, [0, -_x, 0]),
        _NonLinearCageVertex(4, [0, 0, _x]),
        _NonLinearCageVertex(5, [0, 0, -_x]),

        _LinearCageVertex(6, [1, 1, 0], False),
        _LinearCageVertex(7, [1, -1, 0], False),
        _LinearCageVertex(8, [1, 0, 1], False),
        _LinearCageVertex(9, [1, 0, -1], False),
        _LinearCageVertex(10, [-1, 1, 0], False),
        _LinearCageVertex(11, [-1, -1, 0], False),
        _LinearCageVertex(12, [-1, 0, 1], False),
        _LinearCageVertex(13, [-1, 0, -1], False),
        _LinearCageVertex(14, [0, 1, 1], False),
        _LinearCageVertex(15, [0, 1, -1], False),
        _LinearCageVertex(16, [0, -1, 1], False),
        _LinearCageVertex(17, [0, -1, -1], False),
    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[0], _vertex_prototypes[6]),
        Edge(1, _vertex_prototypes[0], _vertex_prototypes[7]),
        Edge(2, _vertex_prototypes[0], _vertex_prototypes[8]),
        Edge(3, _vertex_prototypes[0], _vertex_prototypes[9]),

        Edge(4, _vertex_prototypes[1], _vertex_prototypes[6]),
        Edge(5, _vertex_prototypes[1], _vertex_prototypes[10]),
        Edge(6, _vertex_prototypes[1], _vertex_prototypes[14]),
        Edge(7, _vertex_prototypes[1], _vertex_prototypes[15]),

        Edge(8, _vertex_prototypes[2], _vertex_prototypes[10]),
        Edge(9, _vertex_prototypes[2], _vertex_prototypes[11]),
        Edge(10, _vertex_prototypes[2], _vertex_prototypes[12]),
        Edge(11, _vertex_prototypes[2], _vertex_prototypes[13]),

        Edge(12, _vertex_prototypes[3], _vertex_prototypes[7]),
        Edge(13, _vertex_prototypes[3], _vertex_prototypes[11]),
        Edge(14, _vertex_prototypes[3], _vertex_prototypes[16]),
        Edge(15, _vertex_prototypes[3], _vertex_prototypes[17]),

        Edge(16, _vertex_prototypes[4], _vertex_prototypes[8]),
        Edge(17, _vertex_prototypes[4], _vertex_prototypes[12]),
        Edge(18, _vertex_prototypes[4], _vertex_prototypes[14]),
        Edge(19, _vertex_prototypes[4], _vertex_prototypes[16]),

        Edge(20, _vertex_prototypes[5], _vertex_prototypes[9]),
        Edge(21, _vertex_prototypes[5], _vertex_prototypes[13]),
        Edge(22, _vertex_prototypes[5], _vertex_prototypes[15]),
        Edge(23, _vertex_prototypes[5], _vertex_prototypes[17]),
    )

    _num_windows = 8
    _num_window_types = 1



"""
M8L6 Cube
=========

"""

from ..cage import Cage
from ..vertices import _NonLinearCageVertex
from ...topology_graph import Edge


class M8L6Cube(Cage):
    """
    Represents a cage topology graph.

    Metal building blocks with three functional groups are
    required for this topology.

    Ligand building blocks with four functional groups are required for
    this topology.

    When using a :class:`dict` for initialization, a
    :class:`.BuildingBlock` needs to be assigned to each of the
    following numbers:

        | metals (3 functional groups): 0 to 7
        | ligands (4 functional groups): 8 to 13

    See :class:`.Cage` for more details and examples.

    """

    _vertex_prototypes = (
        _NonLinearCageVertex(
            id=0,
            position=[1, 1, 1],
            use_neighbor_placement=False,
        ),
        _NonLinearCageVertex(
            id=1,
            position=[1, -1, 1],
            use_neighbor_placement=False,
        ),
        _NonLinearCageVertex(
            id=2,
            position=[-1, -1, 1],
            use_neighbor_placement=False,
        ),
        _NonLinearCageVertex(
            id=3,
            position=[-1, 1, 1],
            use_neighbor_placement=False,
        ),
        _NonLinearCageVertex(
            id=4,
            position=[1, 1, -1],
            use_neighbor_placement=False,
        ),
        _NonLinearCageVertex(
            id=5,
            position=[1, -1, -1],
            use_neighbor_placement=False,
        ),
        _NonLinearCageVertex(
            id=6,
            position=[-1, -1, -1],
            use_neighbor_placement=False,
        ),
        _NonLinearCageVertex(
            id=7,
            position=[-1, 1, -1],
            use_neighbor_placement=False,
        ),
    )

    _vertex_prototypes = (
        *_vertex_prototypes,

        _NonLinearCageVertex(
            id=8,
            position=[0, 0, 1],
            use_neighbor_placement=False,
        ),
        _NonLinearCageVertex(
            id=9,
            position=[1, 0, 0],
            use_neighbor_placement=False,
        ),
        _NonLinearCageVertex(
            id=10,
            position=[0, 1, 0],
            use_neighbor_placement=False,
        ),
        _NonLinearCageVertex(
            id=11,
            position=[-1, 0, 0],
            use_neighbor_placement=False,
        ),
        _NonLinearCageVertex(
            id=12,
            position=[0, 0, -1],
            use_neighbor_placement=False,
        ),
        _NonLinearCageVertex(
            id=13,
            position=[0, -1, 0],
            use_neighbor_placement=False,
        ),
    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[0], _vertex_prototypes[8]),
        Edge(1, _vertex_prototypes[1], _vertex_prototypes[8]),
        Edge(2, _vertex_prototypes[2], _vertex_prototypes[8]),
        Edge(3, _vertex_prototypes[3], _vertex_prototypes[8]),

        Edge(4, _vertex_prototypes[4], _vertex_prototypes[9]),
        Edge(5, _vertex_prototypes[5], _vertex_prototypes[9]),
        Edge(6, _vertex_prototypes[1], _vertex_prototypes[9]),
        Edge(7, _vertex_prototypes[0], _vertex_prototypes[9]),

        Edge(8, _vertex_prototypes[4], _vertex_prototypes[10]),
        Edge(9, _vertex_prototypes[0], _vertex_prototypes[10]),
        Edge(10, _vertex_prototypes[3], _vertex_prototypes[10]),
        Edge(11, _vertex_prototypes[7], _vertex_prototypes[10]),

        Edge(12, _vertex_prototypes[3], _vertex_prototypes[11]),
        Edge(13, _vertex_prototypes[2], _vertex_prototypes[11]),
        Edge(14, _vertex_prototypes[6], _vertex_prototypes[11]),
        Edge(15, _vertex_prototypes[7], _vertex_prototypes[11]),

        Edge(16, _vertex_prototypes[5], _vertex_prototypes[12]),
        Edge(17, _vertex_prototypes[4], _vertex_prototypes[12]),
        Edge(18, _vertex_prototypes[7], _vertex_prototypes[12]),
        Edge(19, _vertex_prototypes[6], _vertex_prototypes[12]),

        Edge(20, _vertex_prototypes[1], _vertex_prototypes[13]),
        Edge(21, _vertex_prototypes[5], _vertex_prototypes[13]),
        Edge(22, _vertex_prototypes[6], _vertex_prototypes[13]),
        Edge(23, _vertex_prototypes[2], _vertex_prototypes[13]),
    )

    _num_windows = 4
    _num_window_types = 1



"""
M4L6 Tetrahedron with Spacer
============================

"""

import numpy as np

from ..cage import Cage
from ..vertices import _NonLinearCageVertex, _LinearCageVertex
from ...topology_graph import Edge


class M4L6TetrahedronSpacer(Cage):
    """
    Represents a cage topology graph.

    This topology places a ditopic spacer between the vertices of the
    tetrahderon.

    Metal-based building blocks with three functional groups are
    required for this topology.

    Linker building blocks with two functional groups are required for
    this topology.

    When using a :class:`dict` for initialization, a
    :class:`.BuildingBlock` needs to be assigned to each of the
    following numbers:

        | metals: (0, 1, 2, 3)
        | linkers: (4, 5, 6, 7, 8, 9)

    Examples
    --------
    *Building Metal-Organic Tetrahedron*

    Many metal-organic cages are built using a process called
    subcomponent self-assembly, which is a complex chemical process
    that occurs in solution. In :class:`.Cage`, an example is
    provided of an alchemical approach using
    :class:`.M4L6TetrahedronSpacer` to
    build these types of cages. It is alchemical because the bonds
    formed during construction are not the same as the experimental
    reaction. Instead of forming bonds at the metal centre, we create
    bonds between disconnected ligands.

    The :class:`.M4L6TetrahedronSpacer` topology is provided for cases
    where the linkers cannot be disconnected symmetrically. However,
    in the case that the linker can be disconnected
    in a symmetrical fashion, the
    :class:`.M4L6Tetrahedron` topology can be used.

    """

    _vertex_prototypes = (
        _NonLinearCageVertex(0, [0, 0, np.sqrt(6)/2]),
        _NonLinearCageVertex(1, [-1, -np.sqrt(3)/3, -np.sqrt(6)/6]),
        _NonLinearCageVertex(2, [1, -np.sqrt(3)/3, -np.sqrt(6)/6]),
        _NonLinearCageVertex(3, [0, 2*np.sqrt(3)/3, -np.sqrt(6)/6]),
    )

    _vertex_prototypes = (
        *_vertex_prototypes,

        _LinearCageVertex.init_at_center(
            id=4,
            vertices=(_vertex_prototypes[0], _vertex_prototypes[1]),
        ),
        _LinearCageVertex.init_at_center(
            id=5,
            vertices=(_vertex_prototypes[0], _vertex_prototypes[2]),
        ),
        _LinearCageVertex.init_at_center(
            id=6,
            vertices=(_vertex_prototypes[0], _vertex_prototypes[3]),
        ),
        _LinearCageVertex.init_at_center(
            id=7,
            vertices=(_vertex_prototypes[1], _vertex_prototypes[2]),
        ),
        _LinearCageVertex.init_at_center(
            id=8,
            vertices=(_vertex_prototypes[1], _vertex_prototypes[3]),
        ),
        _LinearCageVertex.init_at_center(
            id=9,
            vertices=(_vertex_prototypes[2], _vertex_prototypes[3]),
        ),
    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[0], _vertex_prototypes[4]),
        Edge(1, _vertex_prototypes[0], _vertex_prototypes[5]),
        Edge(2, _vertex_prototypes[0], _vertex_prototypes[6]),
        Edge(3, _vertex_prototypes[1], _vertex_prototypes[4]),
        Edge(4, _vertex_prototypes[1], _vertex_prototypes[7]),
        Edge(5, _vertex_prototypes[1], _vertex_prototypes[8]),
        Edge(6, _vertex_prototypes[2], _vertex_prototypes[5]),
        Edge(7, _vertex_prototypes[2], _vertex_prototypes[7]),
        Edge(8, _vertex_prototypes[2], _vertex_prototypes[9]),
        Edge(9, _vertex_prototypes[3], _vertex_prototypes[6]),
        Edge(10, _vertex_prototypes[3], _vertex_prototypes[8]),
        Edge(11, _vertex_prototypes[3], _vertex_prototypes[9]),
    )

    _num_windows = 4
    _num_window_types = 1



"""
Bond
====

"""


class Bond:
    """
    Represents an atomic bond.

    """

    def __init__(self, atom1, atom2, order, periodicity=(0, 0, 0)):
        """
        Initialize a :class:`Bond`.

        Parameters
        ----------
        atom1 : :class:`.Atom`
            The first atom in the bond.

        atom2 : :class:`.Atom`
            The second atom in the bond.

        order : :class:`int`
            The bond order.

        periodicity : :class:`tuple` of :class:`int`, optional
            The directions across which the bond is periodic. For
            example, ``(1, 0, -1)`` means that when going from
            `atom1` to `atom2` the bond is
            periodic across the x axis in the positive direction, is
            not periodic across the y axis and is periodic across the z
            axis in the negative direction.

        """

        self._atom1 = atom1
        self._atom2 = atom2
        self._order = order
        self._periodicity = periodicity

    def get_atom1(self):
        """
        Get the first atom of the bond.

        Returns
        -------
        :class:`.Atom`
            The first atom of the bond.

        """

        return self._atom1

    def get_atom2(self):
        """
        Get the second atom of the bond.

        Returns
        -------
        :class:`.Atom`
            The second atom of the bond.

        """

        return self._atom2

    def get_order(self):
        """
        Get the bond order of the bond.

        Returns
        -------
        :class:`int`
            The bond order.

        """

        return self._order

    def get_periodicity(self):
        """
        Get the periodicity of the bond.

        Returns
        -------
        :class:`tuple` of :class:`int`
            The directions across which the bond is periodic. For
            example, ``(1, 0, -1)`` means that when going from
            `atom1` to `atom2` the bond is
            periodic across the x axis in the positive direction, is
            not periodic across the y axis and is periodic across the z
            axis in the negative direction.

        """

        return self._periodicity

    def clone(self):
        """
        Return a clone.

        Returns
        -------
        :class:`.Bond`
            The clone.

        """

        clone = self.__class__.__new__(self.__class__)
        Bond.__init__(
            self=clone,
            atom1=self._atom1,
            atom2=self._atom2,
            order=self._order,
            periodicity=self._periodicity,
        )
        return clone

    def _with_atoms(self, atom_map):
        """
        Modify the bond.

        """

        self._atom1 = atom_map.get(self._atom1.get_id(), self._atom1)
        self._atom2 = atom_map.get(self._atom2.get_id(), self._atom2)
        return self

    def with_atoms(self, atom_map):
        """
        Return a clone holding different atoms.

        Parameters
        ----------
        atom_map : :class:`dict`
            Maps the id of an atom in the bond to the new atom the
            clone should hold. If the id of an atom in the bond is not
            found in `atom_map`, the atom will not be replaced in the
            clone.

        Returns
        -------
        :class:`.Bond`
            The clone.

        Examples
        --------
        .. code-block:: python

            import stk
            bond = stk.Bond(stk.C(0), stk.C(12), 2)
            # Replace C(0) with H(13) but keep C(12).
            clone = bond.with_atoms({0: stk.H(13)})

        """

        return self.clone()._with_atoms(atom_map)

    def is_periodic(self):
        """
        Return ``True`` if the bond is periodic.

        Returns
        -------
        :class:`bool`
            ``True`` if the bond is periodic.

        """

        return any(direction != 0 for direction in self._periodicity)

    def __repr__(self):
        if isinstance(self._order, float) and self._order.is_integer():
            self._order = int(self._order)

        periodicity = (
            f', periodicity={self._periodicity}'
            if self.is_periodic()
            else ''
        )

        cls_name = self.__class__.__name__
        return (
            f'{cls_name}({self._atom1!r}, {self._atom2!r}, '
            f'{self._order}{periodicity})'
        )

    def __str__(self):
        return repr(self)



"""
Linear
======

"""

import numpy as np

from .vertices import _HeadVertex, _TailVertex, _LinearVertex
from ...topology_graph import TopologyGraph, Edge
from ....reactions import GenericReactionFactory


class Linear(TopologyGraph):
    """
    Represents a linear polymer topology graph.

    Examples
    --------
    *Construction*

    Linear polymers require building blocks with two functional groups

    .. code-block:: python

        import stk

        bb1 = stk.BuildingBlock('NCCN', [stk.PrimaryAminoFactory()])
        bb2 = stk.BuildingBlock('O=CCC=O', [stk.AldehydeFactory()])
        polymer = stk.ConstructedMolecule(
            topology_graph=stk.polymer.Linear(
                building_blocks=(bb1, bb2),
                repeating_unit='AB',
                num_repeating_units=12,
            ),
        )

    *Construction with Capping Units*

    Building blocks with a single functional group can
    also be provided as capping units

    .. code-block:: python

        bb3 = stk.BuildingBlock('CCN', [stk.PrimaryAminoFactory()])
        polymer = stk.ConstructedMolecule(
            topology_graph=stk.polymer.Linear(
                building_blocks=(bb1, bb2, bb3),
                repeating_unit='ABABC',
                num_repeating_units=1,
            ),
        )

    *Defining the Orientation of Each Building Block*

    The `orientations` parameter allows the direction of each building
    block along to the chain to be flipped

    .. code-block:: python

        bb4 = stk.BuildingBlock('NCNCCN', [stk.PrimaryAminoFactory()])

        p3 = stk.ConstructedMolecule(
            topology_graph=stk.polymer.Linear(
                building_blocks=(bb2, bb4),
                repeating_unit='AB',
                num_repeating_units=5,
                orientations=(1, 0.5),
            ),
        )

    In the above example, ``bb2`` is guaranteed to be flipped,
    ``bb4`` has a 50% chance of being flipped, each time it is placed
    on a node.

    Note that whether a building block will be flipped or not
    is decided during the initialization of :class:`.Linear`

    .. code-block:: python

        # chain will always construct the same polymer.
        chain = stk.polymer.Linear(
            building_blocks=(bb2, bb4),
            repeating_unit='AB',
            num_repeating_units=5,
            orientations=(0.65, 0.45),
        )
        # p4 and p5 are guaranteed to be the same as they used the same
        # topology graph.
        p4 = stk.ConstructedMolecule(chain)
        p5 = stk.ConstructedMolecule(chain)

        # chain2 may lead to a different polymer than chain, despite
        # being initialized with the same parameters.
        chain2 = stk.polymer.Linear(
            building_blocks=(bb2, bb4),
            repeating_unit='AB',
            num_repeating_units=5,
            orientations=(0.65, 0.45)
        )

        # p6 and p7 are guaranteed to be the same because they used
        # the same topology graph. However, they may be different to
        # p4 and p5.
        p6 = stk.ConstructedMolecule(chain2)
        p7 = stk.ConstructedMolecule(chain2)

    The `random_seed` parameter can be used to get reproducible results

    .. code-block:: python

        # p8 and p9 are guaranteed to be the same, because chain3 and
        # chain4 used the same random seed.

        chain3 = stk.polymer.Linear(
            building_blocks=(bb2, bb4),
            repeating_unit='AB',
            num_repeating_units=5,
            orientations=(0.65, 0.45),
            random_seed=4,
        )
        p8 = stk.ConstructedMolecule(chain3)

        chain4 = stk.polymer.Linear(
            building_blocks=(bb2, bb4),
            repeating_unit='AB',
            num_repeating_units=5,
            orientations=(0.65, 0.45),
            random_seed=4,
        )
        p9 = stk.ConstructedMolecule(chain4)

    *Using Numbers to Define the Repeating Unit*

    The repeating unit can also be specified through the indices of
    the building blocks

    .. code-block:: python

        # p1 and p2 are different ways to write the same thing.
        p1 = stk.ConstructedMolecule(
            topology_graph=stk.polymer.Linear(
                building_blocks=(bb1, bb2, bb3),
                repeating_unit='ACB',
                num_repeating_units=1,
            ),
        )
        p2 = stk.ConstructedMolecule(
            topology_graph=stk.polymer.Linear(
                building_blocks=(bb1, bb2, bb3),
                repeating_unit=(0, 2, 1),
                num_repeating_units=1,
            ),
        )

    """

    def __init__(
        self,
        building_blocks,
        repeating_unit,
        num_repeating_units,
        orientations=None,
        random_seed=None,
        reaction_factory=GenericReactionFactory(),
        num_processes=1,
    ):
        """
        Initialize a :class:`Linear` instance.

        Parameters
        ----------
        building_blocks : :class:`tuple` of :class:`.BuildingBlock`
            The building blocks of the polymer.

        repeating_unit : :class:`str` or :class:`tuple` of :class:`int`
            A string specifying the repeating unit of the polymer.
            For example, ``'AB'`` or ``'ABB'``. The first building
            block passed to `building_blocks` is ``'A'`` and so on.

            The repeating unit can also be specified by the indices of
            `building_blocks`, for example ``'ABB'`` can be
            written as ``(0, 1, 1)``.

        num_repeating_units : :class:`int`
            The number of repeating units which are used to make the
            polymer.

        orientations : :class:`tuple` of :class:`float`, optional
            For each character in the repeating unit, a value
            between ``0`` and ``1`` (both inclusive) must be given in
            a :class:`tuple`. It indicates the probability that each
            monomer will have its orientation along the chain flipped.
            If ``0`` then the monomer is guaranteed not to flip. If
            ``1`` it is guaranteed to flip. This allows the user to
            create head-to-head or head-to-tail chains, as well as
            chain with a preference for head-to-head or head-to-tail if
            a number between ``0`` and ``1`` is chosen. If ``None``
            then ``0`` is picked in all cases.

            It is also possible to supply an orientation for every
            vertex in the final topology graph. In this case, the
            length of `orientations` must be equal to
            ``len(repeating_unit)*num_repeating_units``.

        random_seed : :class:`int`, optional
            The random seed to use when choosing random orientations.

        reaction_factory : :class:`.ReactionFactory`, optional
            The factory to use for creating reactions between
            functional groups of building blocks.

        num_processes : :class:`int`, optional
            The number of parallel processes to create during
            :meth:`construct`.

        Raises
        ------
        :class:`ValueError`
            If the length of `orientations` is not equal in length to
            `repeating_unit` or to the total number of vertices.

        """

        if orientations is None:
            orientations = tuple(
                0. for i in range(len(repeating_unit))
            )

        if len(orientations) == len(repeating_unit):
            orientations = orientations*num_repeating_units

        polymer_length = len(repeating_unit)*num_repeating_units
        if len(orientations) != polymer_length:
            raise ValueError(
                'The length of orientations must match either '
                'the length of repeating_unit or the '
                'total number of vertices.'
            )

        generator = np.random.RandomState(random_seed)

        # Keep these for __repr__.
        self._repeating_unit = self._normalize_repeating_unit(
            repeating_unit=repeating_unit
        )
        self._num_repeating_units = num_repeating_units

        head, *body, tail = orientations
        choices = [True, False]
        vertices = [
            _HeadVertex(
                id=0,
                position=np.array([0, 0, 0]),
                flip=generator.choice(choices, p=[head, 1-head])
            ),
        ]
        edges = []
        for i, p in enumerate(body, 1):
            flip = generator.choice(choices, p=[p, 1-p])
            vertices.append(
                _LinearVertex(i, np.array([i, 0, 0]), flip)
            )
            edges.append(Edge(len(edges), vertices[i-1], vertices[i]))

        vertices.append(
            _TailVertex(
                id=len(vertices),
                position=np.array([len(vertices), 0, 0]),
                flip=generator.choice(choices, p=[tail, 1-tail])),
        )

        # Save the chosen orientations for __repr__.
        self._orientations = tuple(int(v.get_flip()) for v in vertices)

        edges.append(Edge(len(edges), vertices[-2], vertices[-1]))

        super().__init__(
            building_block_vertices=self._get_building_block_vertices(
                building_blocks=building_blocks,
                vertices=vertices,
            ),
            edges=tuple(edges),
            reaction_factory=reaction_factory,
            construction_stages=(),
            num_processes=num_processes,
        )

    def clone(self):
        clone = super().clone()
        clone._repeating_unit = self._repeating_unit
        clone._num_repeating_units = self._num_repeating_units
        clone._orientations = self._orientations
        return clone

    @staticmethod
    def _normalize_repeating_unit(repeating_unit):
        if isinstance(repeating_unit, tuple):
            return repeating_unit

        base = ord('A')
        return tuple(ord(letter)-base for letter in repeating_unit)

    def _get_building_block_vertices(self, building_blocks, vertices):
        polymer = self._repeating_unit*self._num_repeating_units
        building_block_vertices = {}
        for bb_index, vertex in zip(polymer, vertices):
            bb = building_blocks[bb_index]
            building_block_vertices[bb] = (
                building_block_vertices.get(bb, [])
            )
            building_block_vertices[bb].append(vertex)
        return building_block_vertices

    def _get_scale(self, building_block_vertices):
        return max(
            bb.get_maximum_diameter()
            for bb in building_block_vertices
        )

    def __repr__(self):
        return (
            f'polymer.Linear({self._repeating_unit!r}, '
            f'{self._num_repeating_units!r}, '
            f'{self._orientations!r})'
        )



"""
Pdb Writer
==========

"""


class PdbWriter:
    """
    A writer class for ``.pdb`` files.

    Examples
    --------
    *Writing to a File with a Unit Cell*

    This writer can write to a file with the unit
    cell included for periodic molecules. Note that this always assumes
    P1 space group.

    .. code-block:: python

        import stk

        bb1 = stk.BuildingBlock('BrCCBr', [stk.BromoFactory()])
        bb2 = stk.BuildingBlock('BrCC(CBr)CBr', [stk.BromoFactory()])
        topology_graph = stk.cof.PeriodicHoneycomb(
            building_blocks=(bb1, bb2),
            lattice_size=(3, 3, 1),
        )
        cof = stk.ConstructedMolecule(topology_graph)
        writer = stk.PdbWriter()
        writer.write(
            molecule=cof,
            file='cof.pdb',
            periodic_info=topology_graph.get_periodic_info()
        )

    """

    def _write_content(self, molecule, atom_ids, periodic_info=None):

        if atom_ids is None:
            atom_ids = range(molecule.get_num_atoms())
        elif isinstance(atom_ids, int):
            atom_ids = (atom_ids, )

        content = []
        if periodic_info is not None:
            # Input unit cell information.
            a = periodic_info.get_a()
            b = periodic_info.get_b()
            c = periodic_info.get_c()
            alpha = periodic_info.get_alpha()
            beta = periodic_info.get_beta()
            gamma = periodic_info.get_gamma()
            content.append(
                f'CRYST1 {a:>8.3f} {b:>8.3f} {c:>8.3f}'
                f' {alpha:>6.2f} {beta:>6.2f} {gamma:>6.2f} '
                f'P 1\n'
            )

        atom_counts = {}
        hetatm = 'HETATM'
        alt_loc = ''
        res_name = 'UNL'
        chain_id = ''
        res_seq = '1'
        i_code = ''
        occupancy = '1.00'
        temp_factor = '0.00'

        coords = molecule.get_position_matrix()
        # This set will be used by bonds.
        atoms = set()
        for atom_id in atom_ids:
            atoms.add(atom_id)
            atom, = molecule.get_atoms(atom_ids=atom_id)
            serial = atom_id+1
            element = atom.__class__.__name__
            charge = atom.get_charge()
            atom_counts[element] = atom_counts.get(element, 0) + 1
            name = f'{element}{atom_counts[element]}'
            # Make sure the coords are no more than 8 columns wide
            # each.
            x, y, z = (i for i in coords[atom_id])

            content.append(
                f'{hetatm:<6}{serial:>5} {name:<4}'
                f'{alt_loc:<1}{res_name:<3} {chain_id:<1}'
                f'{res_seq:>4}{i_code:<1}   '
                f' {x:>7.3f} {y:>7.3f} {z:>7.3f}'
                f'{occupancy:>6}{temp_factor:>6}          '
                f'{element:>2}{charge:>2}\n'
            )

        conect = 'CONECT'
        for bond in molecule.get_bonds():
            a1 = bond.get_atom1().get_id()
            a2 = bond.get_atom2().get_id()
            if a1 in atoms and a2 in atoms:
                content.append(
                    f'{conect:<6}{a1+1:>5}{a2+1:>5}               \n'
                )

        content.append('END\n')

        return content

    def to_string(
        self,
        molecule,
        atom_ids=None,
        periodic_info=None
    ):
        """
        Get a ``.pdb`` file format string of `molecule`.

        Parameters
        ----------
        molecule : :class:`.Molecule`
            Molecule to write to ``.pdb`` format.

        atom_ids : :class:`iterable` of :class:`int`
            The atom ids of atoms to write. Can be a single
            :class:`int`, if a single atom is to be used, or ``None``,
            if all atoms are to be used.

        periodic_info : :class:`.PeriodicInfo`
            Information about the periodic cell.

        Returns
        -------
        :class:`string`
            A string holding the content of a ``.pdf`` file.

        """

        content = self._write_content(
            molecule=molecule,
            atom_ids=atom_ids,
            periodic_info=periodic_info,
        )

        return ''.join(content)

    def write(
        self,
        molecule,
        path,
        atom_ids=None,
        periodic_info=None
    ):
        """
        Write `molecule` to ``.pdb`` file format.

        Parameters
        ----------
        molecule : :class:`.Molecule`
            Molecule to write to ``.pdb`` format.

        path : :class:`str`
            The full path to the file being written.

        atom_ids : :class:`iterable` of :class:`int`
            The atom ids of atoms to write. Can be a single
            :class:`int`, if a single atom is to be used, or ``None``,
            if all atoms are to be used.

        periodic_info : :class:`.PeriodicInfo`
            Information about the periodic cell.

        Returns
        -------
        None : :class:`NoneType`
            A file is written.

        """

        content = self._write_content(
            molecule=molecule,
            atom_ids=atom_ids,
            periodic_info=periodic_info,
        )

        with open(path, 'w') as f:
            f.write(''.join(content))



"""
Genetic Recombination
=====================

"""

import itertools as it
from collections import defaultdict

from .crosser import MoleculeCrosser
from ...records import CrossoverRecord
from ....molecule_records import MoleculeRecord


class GeneticRecombination(MoleculeCrosser):
    """
    Recombine building blocks using biological systems as a model.

    Overall, this crosser mimics how animals and plants inherit
    DNA from their parents, except generalized to work with any
    number of parents. First it is worth discussing some
    terminology. A gene is a the smallest packet of genetic
    information. In animals, each gene can have multiple alleles.
    For example, there is a gene for hair color, and individual
    alleles for black, red, brown, etc. hair. This means that every
    person has a gene for hair color, but a person with black hair
    will have the black hair allele and a person with red hair will
    have the red hair allele. When two parents produce an
    offspring, the offspring will have a hair color gene and will
    inherit the allele of one of the parents at random. Therefore,
    if you have two parents, one with black hair and one with red
    hair, the offspring will either have black or red hair,
    depending on which allele they inherit.

    In an :mod:`.stk` :class:`.ConstructedMolecule`, each building
    block represents an allele. The question is, which gene is each
    building block an allele of? To answer that, let's first construct
    a couple of building block molecules

    .. code-block:: python

        bb1 = stk.BuildingBlock(
            smiles='NCC(N)CN',
            functional_groups=[stk.PrimaryAminoFactory()],
        )
        bb2 = stk.BuildingBlock('O=CCC=O', [stk.AldehydeFactory()])
        bb3 = stk.BuildingBlock(
            smiles='O=CCNC(C=O)C=O',
            functional_groups=[stk.AldehydeFactory()],
        )
        bb4 = stk.BuildingBlock(
            smiles='NCOCN',
            functional_groups=[stk.PrimaryAminoFactory()],
        )

    We can define a function which analyzes a building block
    molecule and returns the gene it belongs to, for example

    .. code-block:: python

        def get_gene(building_block):
            fg, = building_block.get_functional_groups(0)
            return fg.__class__

    Here, we can see that the gene, to which each building block
    molecule belongs, is given by the class of its first functional
    group. Therefore there is an :class:`.PrimaryAmino` gene,
    which has two alleles ``bb1`` and ``bb4``, and there is an
    :class:`.Aldehyde` gene, which has two alleles ``bb2`` and ``bb3``.

    Alternatively, we could have defined a function such as

    .. code-block:: python

        def get_gene(building_block):
            return building_block.get_num_functional_groups()

    Now we can see that we end up with the gene called
    ``3``, which has two alleles ``bb1`` and ``bb3``,
    and a second gene called ``2``, which has the alleles ``bb2`` and
    ``bb4``.

    To produce offspring molecules, this class categorizes
    each building block of the parent molecules into genes using
    the `get_gene` parameter. Then, to generate a single offspring, it
    picks a building block for every gene. The picked
    building blocks are used to construct the offspring. The
    topology graph of the offspring is one of the parent's.
    For obvious reasons, this approach works with any number of
    parents.

    Examples
    --------
    *Crossing Constructed Molecules*

    Note that any number of parents can be used for the crossover

    .. code-block:: python

        import stk

        # Create the molecule records which will be crossed.

        bb1 = stk.BuildingBlock('NCCN', [stk.PrimaryAminoFactory()])
        bb2 = stk.BuildingBlock('O=CCCCC=O', [stk.AldehydeFactory()])
        graph1 = stk.polymer.Linear((bb1, bb2), 'AB', 2)
        polymer1  = stk.ConstructedMolecule(graph1)
        record1 = stk.MoleculeRecord(polymer1, graph1)

        bb3 = stk.BuildingBlock('NCCCN', [stk.PrimaryAminoFactory()])
        bb4 = stk.BuildingBlock(
            smiles='O=C[Si]CCC=O',
            functional_groups=[stk.AldehydeFactory()],
        )
        graph2 = stk.polymer.Linear((bb3, bb4), 'AB', 2)
        polymer2  = stk.ConstructedMolecule(graph2)
        record2 = stk.MoleculeRecord(polymer2, graph2)

        # Create the crosser.

        def get_functional_group_type(building_block):
            fg, = building_block.get_functional_groups(0)
            return fg.__class__

        recombination = stk.GeneticRecombination(
            get_gene=get_functional_group_type,
        )

        # Get the offspring molecules.

        cohort1 = tuple(recombination.cross(
            records=(record1, record2),
        ))

    """

    def __init__(
        self,
        get_gene,
        name='GeneticRecombination',
    ):
        """
        Initialize a :class:`GeneticRecombination` instance.

        Parameters
        ----------
        get_gene : :class:`callable`
            A :class:`callable`, which takes a :class:`.BuildingBlock`
            object and returns its gene. To produce an offspring, one
            of the building blocks from each gene is picked.

        name : :class:`str`, optional
            A name to identify the crosser instance.

        """

        self._get_gene = get_gene
        self._name = name

    def cross(self, records):
        topology_graphs = (
            record.get_topology_graph() for record in records
        )
        for topology_graph, alleles in it.product(
            topology_graphs,
            self._get_alleles(records),
        ):

            def get_replacement(building_block):
                gene = self._get_gene(building_block)
                return next(
                    allele for allele in alleles
                    if self._get_gene(allele) == gene
                )

            topology_graph = topology_graph.with_building_blocks(
                building_block_map={
                    building_block: get_replacement(building_block)
                    for building_block
                    in topology_graph.get_building_blocks()
                },
            )
            yield CrossoverRecord(
                molecule_record=MoleculeRecord(
                    topology_graph=topology_graph,
                ),
                crosser_name=self._name,
            )

    def _get_alleles(self, records):
        """
        Yield every possible combination of alleles.

        """

        genes = defaultdict(list)
        topology_graphs = (
            record.get_topology_graph() for record in records
        )
        for topology_graph in topology_graphs:
            for allele in topology_graph.get_building_blocks():
                genes[self._get_gene(allele)].append(allele)
        return it.product(*genes.values())
